"""Training/evaluation utilities module.

This module contains utilities and tools used to instantiate training sessions.
"""

import logging
from typing import AnyStr, Optional  # noqa: F401

import thelper.utils

logger = logging.getLogger(__name__)


def create_trainer(session_name,    # type: AnyStr
                   save_dir,        # type: AnyStr
                   config,          # type: thelper.typedefs.ConfigDict
                   model,           # type: thelper.typedefs.ModelType
                   loaders,         # type: thelper.typedefs.MultiLoaderType
                   ckptdata=None    # type: Optional[thelper.typedefs.CheckpointContentType]
                   ):               # type: (...) -> thelper.train.Trainer
    """Instantiates the trainer object based on the type contained in the config dictionary.

    The trainer type is expected to be in the configuration dictionary's `trainer` field, under the `type` key. For more
    information on the configuration, refer to :class:`thelper.train.trainers.Trainer`. The instantiated type must be
    compatible with the constructor signature of :class:`thelper.train.trainers.Trainer`. The object's constructor will
    be given the full config dictionary and the checkpoint data for resuming the session (if available).

    Args:
        session_name: name of the training session used for printing and to create internal tensorboardX directories.
        save_dir: path to the session directory where logs and checkpoints will be saved.
        config: full configuration dictionary that will be parsed for trainer parameters and saved in checkpoints.
        model: model to train/evaluate; should be compatible with :class:`thelper.nn.utils.Module`.
        loaders: a tuple containing the training/validation/test data loaders (a loader can be ``None`` if empty).
        ckptdata: raw checkpoint to parse data from when resuming a session (if ``None``, will start from scratch).

    Returns:
        The fully-constructed trainer object, ready to begin model training/evaluation.

    .. seealso::
        | :class:`thelper.train.trainers.Trainer`

    """
    if "trainer" not in config or not config["trainer"]:
        raise AssertionError("config missing 'trainer' field")
    trainer_config = config["trainer"]
    if "type" not in trainer_config or not trainer_config["type"]:
        raise AssertionError("trainer config missing 'type' field")
    trainer_type = thelper.utils.import_class(trainer_config["type"])
    return trainer_type(session_name, save_dir, model, loaders, config, ckptdata=ckptdata)


# noinspection PyUnusedLocal
def _draw_minibatch_wrapper(sample,         # type: thelper.typedefs.SampleType
                            task,           # type: thelper.tasks.utils.Task
                            pred,           # type: thelper.typedefs.PredictionType
                            iter_idx,       # type: int
                            max_iters,      # type: int
                            epoch_idx,      # type: int
                            max_epochs,     # type: int
                            ):              # type: (...) -> None

    """Wrapper to :func:`thelper.utils.draw_minibatch` used as a callback entrypoint for trainers."""
    thelper.utils.draw_minibatch(sample, task, preds=pred, ch_transpose=True, flip_bgr=False, block=False)



"""Neural network utility functions and classes.

This module contains base interfaces and utility functions used to define and instantiate
neural network models.
"""

import inspect
import logging
import os
from abc import abstractmethod

import numpy as np
import torch
import torch.nn

import thelper
import thelper.nn
import thelper.tasks
import thelper.utils

logger = logging.getLogger(__name__)


def create_model(config, task, save_dir=None, ckptdata=None):
    """Instantiates a model based on a provided task object.

    The configuration must be given as a dictionary object. This dictionary will be parsed for a 'model' field.
    This field is expected to be a dictionary itself. It may then specify a type to instantiate as well as the
    parameters to provide to that class constructor, or a path to a checkpoint from which a model should be loaded.

    All models must derive from :class:`thelper.nn.utils.Module`, or they must be instantiable through
    :class:`thelper.nn.utils.ExternalModule` (or one of its specialized classes). The provided task object will
    be used to make sure that the model has the required input/output layers for the requested objective.

    If checkpoint data is provided by the caller, the weights it contains will be loaded into the returned model.

    Usage examples inside a session configuration file::

        # ...
        # the function will look for a 'model' field in the provided config dict
        "model": {
            # the type provides the class name to instantiate an object from
            "type": "thelper.nn.mobilenet.MobileNetV2",
            # the parameters listed below are passed to the model's constructor
            "params": [
                # ...
            ]
        # ...

    Args:
        config: a session dictionary that provides a 'model' field containing a dictionary.
        task: a task object that will be passed to the model's constructor in order to specialize it. Can be
            ``None`` if a checkpoint is provided, and if the previous task is wanted instead of a new one.
        save_dir: if not ``None``, a log file containing model information will be created there.
        ckptdata: raw checkpoint data loaded via ``torch.load()``; the model will be given its previous state.

    Returns:
        The instantiated model, compatible with the interface of both :class:`thelper.nn.utils.Module`
        and ``torch.nn.Module``.

    .. seealso::
        | :class:`thelper.nn.utils.Module`
        | :class:`thelper.nn.utils.ExternalModule`
        | :class:`thelper.tasks.utils.Task`
    """
    if save_dir is not None:
        modules_logger_path = os.path.join(save_dir, "logs", "modules.log")
        modules_logger_format = logging.Formatter("[%(asctime)s - %(process)s] %(levelname)s : %(message)s")
        modules_logger_fh = logging.FileHandler(modules_logger_path)
        modules_logger_fh.setFormatter(modules_logger_format)
        logger.addHandler(modules_logger_fh)
        logger.info("created modules log for session '%s'" % config["name"])
    logger.debug("loading model")
    if "model" not in config or not config["model"]:
        raise AssertionError("config missing 'model' field")
    model_config = config["model"]
    if "ckptdata" in model_config:
        if ckptdata is not None:
            logger.warning("config asked to reload ckpt from path, but ckpt already loaded from elsewhere")
        else:
            logger.debug("model config asked for an older model to be loaded through a checkpoint")
            if not isinstance(model_config["ckptdata"], str):
                raise AssertionError("unexpected model config ckptdata field type (should be path)")
            map_location = thelper.utils.get_key_def("map_location", model_config, "cpu")
            ckptdata = thelper.utils.load_checkpoint(model_config["ckptdata"], map_location=map_location)
        if "type" in model_config or "params" in model_config:
            logger.warning("should not provide 'type' or 'params' fields in model config if loading a checkpoint")
    new_task, model, model_type, model_params, model_state = None, None, None, None, None
    if ckptdata is not None:
        # if checkpoint available, instantiate old model, load weights, and reconfigure for new task
        if "name" not in ckptdata or not isinstance(ckptdata["name"], str):
            raise AssertionError("invalid checkpoint, cannot reload previous session name")
        new_task = task  # the 'new task' will later be applied to specialize the model, once it is loaded
        if "model" not in ckptdata or not isinstance(ckptdata["model"], (Module, dict)):
            raise AssertionError("invalid checkpoint, cannot reload previous model state")
        if isinstance(ckptdata["model"], Module):
            logger.debug("loading model object directly from session '%s'" % ckptdata["name"])
            model = ckptdata["model"]
            if model.get_name() != ckptdata["model_type"]:
                raise AssertionError("old model type mismatch with ckptdata type")
        elif isinstance(ckptdata["model"], dict):
            logger.debug("loading model type/params from session '%s'" % ckptdata["name"])
            model_state = ckptdata["model"]
            if "task" not in ckptdata or not isinstance(ckptdata["task"], (thelper.tasks.Task, str)):
                raise AssertionError("invalid checkpoint, cannot reload previous model task")
            task = thelper.tasks.create_task(ckptdata["task"]) if isinstance(ckptdata["task"], str) else ckptdata["task"]
            if "model_type" not in ckptdata or not isinstance(ckptdata["model_type"], str):
                raise AssertionError("invalid checkpoint, cannot reload previous model type")
            model_type = thelper.utils.import_class(ckptdata["model_type"])
            if "model_params" not in ckptdata or not isinstance(ckptdata["model_params"], dict):
                raise AssertionError("invalid checkpoint, cannot reload previous model params")
            model_params = ckptdata["model_params"]
            if "config" not in ckptdata or not isinstance(ckptdata["config"], dict):
                raise AssertionError("invalid checkpoint, cannot reload previous session config")
            old_config = ckptdata["config"]
            if "model" not in old_config or not isinstance(old_config["model"], dict):
                raise AssertionError("invalid checkpoint, cannot reload previous model config")
            old_model_config = old_config["model"]
            if "type" in old_model_config and thelper.utils.import_class(old_model_config["type"]) != model_type:
                raise AssertionError("old model config 'type' field mismatch with ckptdata type")
    else:
        if not isinstance(task, thelper.tasks.Task):
            raise AssertionError("bad task type passed to create_model")
        logger.debug("loading model type/params current config")
        if "type" not in model_config or not model_config["type"]:
            raise AssertionError("model config missing 'type' field")
        model_type = thelper.utils.import_class(model_config["type"])
        model_params = thelper.utils.get_key_def("params", model_config, {})
    if model is None:
        # if model not already loaded from checkpoint, instantiate it fully from type/params/task
        if model_type is None or model_params is None:
            raise AssertionError("messed up logic above")
        logger.debug("model_type = %s" % str(model_type))
        logger.debug("model_params = %s" % str(model_params))
        if inspect.isclass(model_type) and issubclass(model_type, thelper.nn.utils.Module):
            model = model_type(task=task, **model_params)
        else:
            if type(task) == thelper.tasks.Classification:
                model = ExternalClassifModule(model_type, task=task, config=model_params)
            else:
                model = ExternalModule(model_type, task=task, config=model_params)
        if model_state is not None:
            logger.debug("loading state dictionary from checkpoint into model")
            model.load_state_dict(model_state)
    if new_task is not None:
        logger.debug("previous model task = %s" % str(model.task))
        logger.debug("refreshing model for new task = %s" % str(new_task))
        model.set_task(new_task)
    if model.config is None:
        model.config = model_params
    if hasattr(model, "summary"):
        model.summary()
    return model


class Module(torch.nn.Module):
    """Model interface used to hold a task object.

    This interface is built on top of ``torch.nn.Module`` and should remain fully compatible with it.

    All models used in the framework should derive from this interface, and therefore expect a task object as
    the first argument of their constructor. Their implementation may decide to ignore this task object when
    building their internal layers, but using it should help specialize the network by specifying e.g. the
    number of classes to support.

    .. seealso::
        | :func:`thelper.nn.utils.create_model`
        | :class:`thelper.tasks.utils.Task`
    """

    def __init__(self, task, config=None):
        """Receives a task object to hold internally for model specialization."""
        super().__init__()
        if task is None or not isinstance(task, thelper.tasks.Task):
            raise AssertionError("task must derive from thelper.tasks.Task")
        self.task = task
        self.config = config

    @abstractmethod
    def forward(self, *input):
        """Transforms an input tensor in order to generate a prediction."""
        raise NotImplementedError

    @abstractmethod
    def set_task(self, task):
        """Adapts the model to support a new task, replacing layers if needed."""
        raise NotImplementedError

    def summary(self):
        """Prints a summary of the model using the ``thelper.nn`` logger."""
        params = filter(lambda p: p.requires_grad, self.parameters())
        count = sum([np.prod(p.size()) for p in params])
        logger.info("module '%s' parameter count: %d" % (self.get_name(), count))
        logger.info(self)

    def get_name(self):
        """Returns the name of this module (by default, its fully qualified class name)."""
        return self.__class__.__module__ + "." + self.__class__.__qualname__


class ExternalModule(Module):
    """Model inteface used to hold a task object for an external implementation.

    This interface is built on top of ``torch.nn.Module`` and should remain fully compatible with it. It is
    automatically used when instantiating a model via :func:`thelper.nn.utils.create_model` that is not derived
    from :class:`thelper.nn.utils.Module`. Its only purpose is to hold the task object, and redirect
    :func:`thelper.nn.utils.Module.forward` to the actual model's transformation function. It can also be
    specialized to automatically adapt some external models after their construction using the knowledge
    contained in the task object.

    .. seealso::
        | :class:`thelper.nn.utils.Module`
        | :class:`thelper.nn.utils.ExternalClassifModule`
        | :func:`thelper.nn.utils.create_model`
        | :class:`thelper.tasks.utils.Task`
    """

    def __init__(self, model_type, task, config=None):
        """Receives a task object to hold internally for model specialization."""
        super().__init__(task=task, config=config)
        logger.info("instantiating external module '%s'..." % str(model_type))
        self.model_type = model_type
        self.model = model_type(**config)
        if not hasattr(self.model, "forward"):
            raise AssertionError("external module must implement 'forward' method")

    def load_state_dict(self, state_dict, strict=True):
        """Loads the state dict of an external model."""
        self.model.load_state_dict(state_dict=state_dict, strict=strict)

    def state_dict(self, destination=None, prefix='', keep_vars=False):
        """Returns the state dict of the external model."""
        return self.model.state_dict(destination=destination, prefix=prefix, keep_vars=keep_vars)

    def forward(self, *input):
        """Transforms an input tensor in order to generate a prediction."""
        return self.model(*input)

    def set_task(self, task):
        """Stores the new task internally.

        Note that since this external module handler is generic, it does not know what to do with the task,
        so it just assumes that the model is already set up. Specialized external module handlers will instead
        attempt to modify the model they wrap.
        """
        if task is None or not isinstance(task, thelper.tasks.Task):
            raise AssertionError("task must derive from thelper.tasks.Task")
        self.task = task

    def summary(self):
        """Prints a summary of the model using the ``thelper.nn`` logger."""
        params = filter(lambda p: p.requires_grad, self.model.parameters())
        count = sum([np.prod(p.size()) for p in params])
        logger.info("module '%s' parameter count: %d" % (self.get_name(), count))
        logger.info(self.model)

    def get_name(self):
        """Returns the name of this module (by default, the fully qualified class name of the external model)."""
        return self.model_type.__module__ + "." + self.model_type.__qualname__


class ExternalClassifModule(ExternalModule):
    """External model interface specialization for classification tasks.

    This interface will try to 'rewire' the last fully connected layer of the models it instantiates to match
    the number of classes to predict defined in the task object.

    .. seealso::
        | :class:`thelper.nn.utils.Module`
        | :class:`thelper.nn.utils.ExternalModule`
        | :func:`thelper.nn.utils.create_model`
        | :class:`thelper.tasks.utils.Task`
    """

    def __init__(self, model_type, task, config=None):
        """Receives a task object to hold internally for model specialization, and tries to rewire the last 'fc' layer."""
        super().__init__(model_type, task, config=config)
        self.nb_classes = None
        self.set_task(task)

    def set_task(self, task):
        """Rewires the last fully connected layer of the wrapped network to fit the given number of classification targets."""
        if type(task) != thelper.tasks.Classification:
            raise AssertionError("task passed to ExternalClassifModule should be 'thelper.tasks.Classification'")
        self.nb_classes = self.task.get_nb_classes()
        import torchvision
        if hasattr(self.model, "fc") and isinstance(self.model.fc, torch.nn.Linear):
            if self.model.fc.out_features != self.nb_classes:
                logger.info("reconnecting fc layer for outputting %d classes..." % self.nb_classes)
                nb_features = self.model.fc.in_features
                self.model.fc = torch.nn.Linear(nb_features, self.nb_classes)
        elif hasattr(self.model, "classifier") and isinstance(self.model.classifier, torch.nn.Linear):
            if self.model.classifier.out_features != self.nb_classes:
                logger.info("reconnecting classifier layer for outputting %d classes..." % self.nb_classes)
                nb_features = self.model.classifier.in_features
                self.model.classifier = torch.nn.Linear(nb_features, self.nb_classes)
        elif isinstance(self.model, torchvision.models.squeezenet.SqueezeNet):
            if self.model.num_classes != self.nb_classes:
                self.model.classifier = torch.nn.Sequential(
                    torch.nn.Dropout(p=0.5),
                    torch.nn.Conv2d(512, self.nb_classes, kernel_size=1),
                    torch.nn.ReLU(inplace=True),
                    torch.nn.AvgPool2d(13, stride=1)
                )
                self.model.num_classes = self.nb_classes
        else:
            raise AssertionError("could not reconnect fully connected layer for new classes")



import torch
import torch.nn
import torch.utils.model_zoo

import thelper.nn
import thelper.nn.coordconv

__all__ = ['ResNet', 'resnet18', 'resnet34', 'resnet50', 'resnet101',
           'resnet152']

model_urls = {
    'resnet18': 'https://download.pytorch.org/models/resnet18-5c106cde.pth',
    'resnet34': 'https://download.pytorch.org/models/resnet34-333f7ec4.pth',
    'resnet50': 'https://download.pytorch.org/models/resnet50-19c8e357.pth',
    'resnet101': 'https://download.pytorch.org/models/resnet101-5d3b4d8f.pth',
    'resnet152': 'https://download.pytorch.org/models/resnet152-b121ed2d.pth',
}


class Module(torch.nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, downsample=None, coordconv=False, radius_channel=True):
        super().__init__()
        self.inplanes = inplanes
        self.planes = planes
        self.stride = stride
        self.downsample = downsample
        self.coordconv = coordconv
        self.radius_channel = radius_channel

    def _make_conv2d(self, *args, **kwargs):
        if self.coordconv:
            return thelper.nn.coordconv.CoordConv2d(*args, radius_channel=self.radius_channel, **kwargs)
        else:
            return torch.nn.Conv2d(*args, **kwargs)


class BasicBlock(Module):

    def __init__(self, inplanes, planes, stride=1, downsample=None, coordconv=False, radius_channel=True):
        super().__init__(inplanes, planes, stride, downsample, coordconv, radius_channel)
        self.conv1 = self._make_conv2d(inplanes, planes, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn1 = torch.nn.BatchNorm2d(planes)
        self.relu = torch.nn.ReLU(inplace=True)
        self.conv2 = self._make_conv2d(planes, planes, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn2 = torch.nn.BatchNorm2d(planes)

    def forward(self, x):
        residual = x
        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)
        out = self.conv2(out)
        out = self.bn2(out)
        if self.downsample is not None:
            residual = self.downsample(x)
        out += residual
        out = self.relu(out)
        return out


class Bottleneck(Module):
    expansion = 4

    def __init__(self, inplanes, planes, stride=1, downsample=None, coordconv=False, radius_channel=True):
        super().__init__(inplanes, planes, stride, downsample, coordconv, radius_channel)
        self.conv1 = self._make_conv2d(inplanes, planes, kernel_size=1, bias=False)
        self.bn1 = torch.nn.BatchNorm2d(planes)
        self.conv2 = self._make_conv2d(planes, planes, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn2 = torch.nn.BatchNorm2d(planes)
        self.conv3 = self._make_conv2d(planes, planes * self.expansion, kernel_size=1, bias=False)
        self.bn3 = torch.nn.BatchNorm2d(planes * self.expansion)
        self.relu = torch.nn.ReLU(inplace=True)

    def forward(self, x):
        residual = x
        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)
        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)
        out = self.conv3(out)
        out = self.bn3(out)
        if self.downsample is not None:
            residual = self.downsample(x)
        out += residual
        out = self.relu(out)
        return out


class SqueezeExcitationLayer(torch.nn.Module):

    def __init__(self, channel, reduction=16):
        super().__init__()
        self.pool = torch.nn.AdaptiveAvgPool2d(1)
        self.fc = torch.nn.Sequential(
            torch.nn.Linear(channel, channel // reduction),
            torch.nn.ReLU(inplace=True),
            torch.nn.Linear(channel // reduction, channel),
            torch.nn.Sigmoid()
        )

    def forward(self, x):
        b, c, _, _ = x.size()
        y = self.pool(x).view(b, c)
        y = self.fc(y).view(b, c, 1, 1)
        return x * y


class SqueezeExcitationBlock(Module):

    def __init__(self, inplanes, planes, stride=1, downsample=None, reduction=16, coordconv=False, radius_channel=True):
        super().__init__(inplanes, planes, stride, downsample, coordconv, radius_channel)
        self.conv1 = self._make_conv2d(inplanes, planes, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn1 = torch.nn.BatchNorm2d(planes)
        self.relu = torch.nn.ReLU(inplace=True)
        self.conv2 = self._make_conv2d(planes, planes, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn2 = torch.nn.BatchNorm2d(planes)
        self.se = SqueezeExcitationLayer(planes, reduction)

    def forward(self, x):
        residual = x
        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)
        out = self.conv2(out)
        out = self.bn2(out)
        out = self.se(out)
        if self.downsample is not None:
            residual = self.downsample(x)
        out += residual
        out = self.relu(out)
        return out


class ResNet(thelper.nn.Module):

    def __init__(self, task, block=BasicBlock, layers=[3, 4, 6, 3], strides=[1, 2, 2, 2], input_channels=3,
                 flexible_input_res=False, pool_size=7, coordconv=False, radius_channel=True):
        super().__init__(task)
        if isinstance(block, str):
            block = thelper.utils.import_class(block)
        if not issubclass(block, Module):
            raise AssertionError("block type must be subclass of thelper.nn.resnet.Module")
        if not isinstance(layers, list) or not isinstance(strides, list):
            raise AssertionError("expected layers/strides to be provided as list of ints")
        if len(layers) != len(strides):
            raise AssertionError("layer/strides length mismatch")
        self.inplanes = 64
        self.coordconv = coordconv
        self.radius_channel = radius_channel
        self.conv1 = self._make_conv2d(in_channels=input_channels, out_channels=self.inplanes,
                                       kernel_size=7, stride=2, padding=3, bias=False)
        self.bn1 = torch.nn.BatchNorm2d(self.inplanes)
        self.relu = torch.nn.ReLU(inplace=True)
        self.maxpool = torch.nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = self._make_layer(block, 64, layers[0], stride=strides[0])
        self.layer2 = self._make_layer(block, 128, layers[1], stride=strides[1])
        self.layer3 = self._make_layer(block, 256, layers[2], stride=strides[2])
        self.layer4 = self._make_layer(block, 512, layers[3], stride=strides[3])
        self.out_features = 512
        self.layer5 = None
        if len(layers) > 4:
            self.layer5 = self._make_layer(block, 1024, layers[4], stride=strides[4])
            self.out_features = 1024
        if flexible_input_res:
            self.avgpool = torch.nn.AdaptiveAvgPool2d(1)
        else:
            if pool_size < 1:
                raise AssertionError("invalid avg pool size for non-flex resolution")
            self.avgpool = torch.nn.AvgPool2d(pool_size, stride=1)
        self.out_features *= block.expansion
        self.fc = torch.nn.Linear(self.out_features, 1000)  # output type/count will be specialized by task after init
        for m in self.modules():
            if isinstance(m, torch.nn.Conv2d):
                torch.nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, thelper.nn.coordconv.CoordConv2d):
                torch.nn.init.kaiming_normal_(m.conv.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, torch.nn.BatchNorm2d):
                torch.nn.init.constant_(m.weight, 1)
                torch.nn.init.constant_(m.bias, 0)
        self.set_task(task)

    def _make_conv2d(self, *args, **kwargs):
        if self.coordconv:
            return thelper.nn.coordconv.CoordConv2d(*args, radius_channel=self.radius_channel, **kwargs)
        else:
            return torch.nn.Conv2d(*args, **kwargs)

    def _make_layer(self, block, planes, blocks, stride):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = torch.nn.Sequential(
                self._make_conv2d(self.inplanes, planes * block.expansion, kernel_size=1, stride=stride, bias=False),
                torch.nn.BatchNorm2d(planes * block.expansion),
            )
        layers = [block(self.inplanes, planes, stride, downsample)]
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes))
        return torch.nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)
        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)
        if self.layer5 is not None:
            x = self.layer5(x)
        x = self.avgpool(x)
        x = x.view(x.size(0), -1)
        x = self.fc(x)
        return x

    def set_task(self, task):
        if isinstance(task, thelper.tasks.Classification):
            num_classes = len(task.get_class_names())
            if self.fc.out_features != num_classes:
                self.fc = torch.nn.Linear(self.out_features, num_classes)
        else:
            raise AssertionError("missing impl for non-classif task type")
        self.task = task


class FCResNet(ResNet):

    def __init__(self, task, ckptdata, map_location="cpu", avgpool_size=0):
        if isinstance(ckptdata, str):
            ckptdata = thelper.utils.load_checkpoint(ckptdata, map_location=map_location)
        model_type = ckptdata["model_type"]
        if model_type != "thelper.nn.resnet.ResNet":
            raise AssertionError("cannot convert non-resnet model to fully conv with this impl")
        model_params = ckptdata["model_params"]
        if isinstance(ckptdata["task"], str):
            old_model_task = thelper.tasks.create_task(ckptdata["task"])
        else:
            old_model_task = ckptdata["task"]
        self.task = None
        self.avgpool_size = avgpool_size
        super().__init__(old_model_task, **model_params)
        self.load_state_dict(ckptdata["model"], strict=False)  # assumes model always stored as weight dict
        self.finallayer = torch.nn.Conv2d(self.out_features, self.fc.out_features, kernel_size=1)
        self.finallayer.weight = torch.nn.Parameter(self.fc.weight.view(self.fc.out_features, self.out_features, 1, 1))
        self.finallayer.bias = torch.nn.Parameter(self.fc.bias)
        self.set_task(task)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)
        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)
        if self.layer5 is not None:
            x = self.layer5(x)
        if self.avgpool_size > 0:
            x = torch.nn.functional.avg_pool2d(x, kernel_size=self.avgpool_size, stride=1)
        x = self.finallayer(x)
        return x

    def set_task(self, task):
        if isinstance(task, (thelper.tasks.Segmentation, thelper.tasks.Classification)):
            num_classes = len(task.get_class_names())
            if self.fc.out_features != num_classes:
                self.fc = torch.nn.Linear(self.out_features, num_classes)
                self.finallayer = torch.nn.Conv2d(self.out_features, num_classes, kernel_size=1)
                self.finallayer.weight = torch.nn.Parameter(self.fc.weight.view(self.fc.out_features, self.out_features, 1, 1))
                self.finallayer.bias = torch.nn.Parameter(self.fc.bias)
        else:
            raise AssertionError("missing impl for non-segm/classif task type")
        self.task = task


def resnet18(pretrained=False, **kwargs):
    """Constructs a ResNet-18 model.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet.
    """
    model = ResNet(BasicBlock, [2, 2, 2, 2], **kwargs)
    if pretrained:
        model.load_state_dict(torch.utils.model_zoo.load_url(model_urls['resnet18']))
    return model


def resnet34(pretrained=False, **kwargs):
    """Constructs a ResNet-34 model.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet.
    """
    model = ResNet(BasicBlock, [3, 4, 6, 3], **kwargs)
    if pretrained:
        model.load_state_dict(torch.utils.model_zoo.load_url(model_urls['resnet34']))
    return model


def resnet50(pretrained=False, **kwargs):
    """Constructs a ResNet-50 model.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet.
    """
    model = ResNet(Bottleneck, [3, 4, 6, 3], **kwargs)
    if pretrained:
        model.load_state_dict(torch.utils.model_zoo.load_url(model_urls['resnet50']))
    return model


def resnet101(pretrained=False, **kwargs):
    """Constructs a ResNet-101 model.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet.
    """
    model = ResNet(Bottleneck, [3, 4, 23, 3], **kwargs)
    if pretrained:
        model.load_state_dict(torch.utils.model_zoo.load_url(model_urls['resnet101']))
    return model


def resnet152(pretrained=False, **kwargs):
    """Constructs a ResNet-152 model.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet.
    """
    model = ResNet(Bottleneck, [3, 8, 36, 3], **kwargs)
    if pretrained:
        model.load_state_dict(torch.utils.model_zoo.load_url(model_urls['resnet152']))
    return model



"""Dataset loaders module.

This module contains a dataset loader specialization used to properly seed samplers and workers.
"""

import copy
import inspect
import logging
import random
import sys
import time
from collections import Counter

import numpy as np
import torch
import torch.utils.data
import torch.utils.data.sampler
import tqdm

import thelper.tasks
import thelper.transforms
import thelper.utils

logger = logging.getLogger(__name__)


class DataLoader(torch.utils.data.DataLoader):
    """Specialized data loader used to load minibatches from a dataset parser.

    This specialization handles the seeding of samplers and workers.

    See ``torch.utils.data.DataLoader`` for more information on attributes/methods.
    """
    def __init__(self, *args, seeds=None, epoch=0, **kwargs):
        super().__init__(*args, worker_init_fn=self._worker_init_fn, **kwargs)
        self.seeds = {}
        if seeds is not None:
            if not isinstance(seeds, dict):
                raise AssertionError("unexpected seed pack type")
            self.seeds = seeds
        if not isinstance(epoch, int) or epoch < 0:
            raise AssertionError("invalid epoch value")
        self.epoch = epoch

    def __iter__(self):
        """Advances the epoch number for the workers initialization function."""
        result = super().__iter__()
        self.epoch += 1
        return result

    def set_epoch(self, epoch=0):
        """Sets the current epoch number in order to offset RNG states for the workers and the sampler."""
        if not isinstance(epoch, int) or epoch < 0:
            raise AssertionError("invalid epoch value")
        self.epoch = epoch
        if self.sampler is not None:
            if hasattr(self.sampler, "set_epoch") and callable(self.sampler.set_epoch):
                self.sampler.set_epoch(self.epoch)
        if hasattr(self, "dataset"):
            if hasattr(self.dataset, "set_epoch") and callable(self.dataset.set_epoch):
                self.dataset.set_epoch(epoch)
            if hasattr(self.dataset, "transforms"):
                if hasattr(self.dataset.transforms, "set_epoch") and callable(self.dataset.transforms.set_epoch):
                    self.dataset.transforms.set_epoch(epoch)

    def _worker_init_fn(self, worker_id):
        """Sets up the RNGs state of each worker based on their unique id and the epoch number."""
        seed_offset = self.num_workers * self.epoch
        if "torch" in self.seeds:
            torch.manual_seed(self.seeds["torch"] + seed_offset + worker_id)
            torch.cuda.manual_seed_all(self.seeds["torch"] + seed_offset + worker_id)
        if "numpy" in self.seeds:
            np.random.seed(self.seeds["numpy"] + seed_offset + worker_id)
        if "random" in self.seeds:
            random.seed(self.seeds["random"] + seed_offset + worker_id)

    @property
    def sample_count(self):
        if self.sampler is not None:
            return len(self.sampler)
        return len(self.dataset)


class _LoaderFactory(object):
    """Factory used for preparing and splitting dataset parsers into usable data loader objects.

    This class is responsible for parsing the parameters contained in the 'loaders' field of a
    configuration dictionary, instantiating the data loaders, and shuffling/splitting the samples.
    An example configuration is presented in :func:`thelper.data.utils.create_loaders`.

    .. seealso::
        | :func:`thelper.data.utils.create_loaders`
        | :func:`thelper.transforms.utils.load_augments`
        | :func:`thelper.transforms.utils.load_transforms`
    """

    def __init__(self, config):
        """Receives and parses the data configuration dictionary."""
        logger.debug("loading data config")
        if not isinstance(config, dict):
            raise AssertionError("input config should be dict")
        default_batch_size = 0
        if "batch_size" in config:
            if any([v in config for v in ["train_batch_size", "valid_batch_size", "test_batch_size"]]):
                raise AssertionError("specifying 'batch_size' overrides all other (loader-specific) values")
            default_batch_size = int(thelper.utils.get_key("batch_size", config))
        self.train_batch_size = thelper.utils.get_key_def("train_batch_size", config, default_batch_size)
        self.valid_batch_size = thelper.utils.get_key_def("valid_batch_size", config, default_batch_size)
        self.test_batch_size = thelper.utils.get_key_def("test_batch_size", config, default_batch_size)
        logger.debug("loaders will use batch sizes:\n  train = %d\n  valid = %d\n  test = %d" %
                     (self.train_batch_size, self.valid_batch_size, self.test_batch_size))
        self.train_scale = thelper.utils.get_key_def("train_scale", config, 1.0)
        self.valid_scale = thelper.utils.get_key_def("valid_scale", config, 1.0)
        self.test_scale = thelper.utils.get_key_def("test_scale", config, 1.0)
        logger.debug("samplers will use scaling factors:\n  train = %f\n  valid = %f\n  test = %f" %
                     (self.train_scale, self.valid_scale, self.test_scale))
        default_collate_fn = torch.utils.data.dataloader.default_collate
        if "collate_fn" in config:
            if any([v in config for v in ["train_collate_fn", "valid_collate_fn", "test_collate_fn"]]):
                raise AssertionError("specifying 'collate_fn' overrides all other (loader-specific) values")
            default_collate_fn = self._get_collate_fn(config["collate_fn"])
        self.train_collate_fn = self._get_collate_fn(thelper.utils.get_key_def("train_collate_fn", config, default_collate_fn))
        self.valid_collate_fn = self._get_collate_fn(thelper.utils.get_key_def("valid_collate_fn", config, default_collate_fn))
        self.test_collate_fn = self._get_collate_fn(thelper.utils.get_key_def("test_collate_fn", config, default_collate_fn))
        self.train_shuffle = thelper.utils.str2bool(thelper.utils.get_key_def(["shuffle", "train_shuffle"], config, True))
        self.valid_shuffle = thelper.utils.str2bool(thelper.utils.get_key_def("valid_shuffle", config, False))
        self.test_shuffle = thelper.utils.str2bool(thelper.utils.get_key_def("test_shuffle", config, False))
        np.random.seed()  # for seed generation below (if needed); will be reseeded afterwards
        test_seed = self._get_seed(["test_seed", "test_split_seed"], config, (int, str))
        valid_seed = self._get_seed(["valid_seed", "valid_split_seed"], config, (int, str))
        torch_seed = self._get_seed(["torch_seed"], config, int)
        numpy_seed = self._get_seed(["numpy_seed"], config, int)
        random_seed = self._get_seed(["random_seed"], config, int)
        torch.manual_seed(torch_seed)
        torch.cuda.manual_seed_all(torch_seed)
        np.random.seed(numpy_seed)
        random.seed(random_seed)
        self.seeds = {
            "test": test_seed,
            "valid": valid_seed,
            "torch": torch_seed,
            "numpy": numpy_seed,
            "random": random_seed
        }
        self.workers = config["workers"] if "workers" in config and config["workers"] >= 0 else 1
        self.pin_memory = thelper.utils.str2bool(config["pin_memory"]) if "pin_memory" in config else False
        self.drop_last = thelper.utils.str2bool(config["drop_last"]) if "drop_last" in config else False
        if self.drop_last:
            logger.debug("loaders will drop last batch if sample count not multiple of batch size")
        self.sampler_type = None
        self.train_sampler, self.valid_sampler, self.test_sampler = None, None, None
        if "sampler" in config:
            sampler_config = config["sampler"]
            if sampler_config:
                if "type" not in sampler_config or not sampler_config["type"]:
                    raise AssertionError("missing 'type' field for sampler config")
                self.sampler_type = thelper.utils.import_class(sampler_config["type"])
                self.sampler_params = thelper.utils.get_key_def(["params", "parameters"], sampler_config, {})
                logger.debug("will use sampler with type '%s' and config : %s" % (str(self.sampler_type), str(self.sampler_params)))
                self.sampler_pass_labels = thelper.utils.str2bool(thelper.utils.get_key_def("pass_labels", sampler_config, False))
                self.sampler_pass_labels_param_name = thelper.utils.get_key_def("pass_labels_param_name", sampler_config, "labels")
                self.train_sampler = thelper.utils.str2bool(thelper.utils.get_key_def("apply_train", sampler_config, True))
                self.valid_sampler = thelper.utils.str2bool(thelper.utils.get_key_def("apply_valid", sampler_config, False))
                self.test_sampler = thelper.utils.str2bool(thelper.utils.get_key_def("apply_test", sampler_config, False))
                logger.debug("global sampler will be applied as: %s" % str([self.train_sampler, self.valid_sampler, self.test_sampler]))
        train_augs_targets = ["augments", "trainvalid_augments", "train_augments"]
        valid_augs_targets = ["augments", "trainvalid_augments", "eval_augments", "validtest_augments", "valid_augments"]
        test_augs_targets = ["augments", "eval_augments", "validtest_augments", "test_augments"]
        self.train_augments, self.train_augments_append = self._get_augments(train_augs_targets, "train", config)
        self.valid_augments, self.valid_augments_append = self._get_augments(valid_augs_targets, "valid", config)
        self.test_augments, self.test_augments_append = self._get_augments(test_augs_targets, "test", config)
        self.base_transforms = None
        if "base_transforms" in config and config["base_transforms"]:
            logger.debug("loading base transforms...")
            self.base_transforms = thelper.transforms.load_transforms(config["base_transforms"])
            if self.base_transforms:
                logger.debug("base transforms: %s" % str(self.base_transforms))
        self.train_split = self._get_ratios_split("train", config)
        self.valid_split = self._get_ratios_split("valid", config)
        self.test_split = self._get_ratios_split("test", config)
        if not self.train_split and not self.valid_split and not self.test_split:
            raise AssertionError("data config must define a split for at least one loader type (train/valid/test)")
        self.total_usage = Counter(self.train_split) + Counter(self.valid_split) + Counter(self.test_split)
        self.skip_split_norm = thelper.utils.str2bool(thelper.utils.get_key_def("skip_split_norm", config, False))
        self.skip_class_balancing = thelper.utils.str2bool(thelper.utils.get_key_def("skip_class_balancing", config, False))
        for name, usage in self.total_usage.items():
            if usage != 1:
                normalize_ratios = None
                if usage < 0:
                    raise AssertionError("ratio should never be negative...")
                elif 0 < usage < 1 and not self.skip_split_norm:
                    query_msg = "dataset split for '%s' has a ratio sum less than 1; do you want to normalize the split?" % name
                    normalize_ratios = thelper.utils.query_yes_no(query_msg, bypass="n")
                if (normalize_ratios or usage > 1) and usage > 0:
                    if usage > 1:
                        logger.warning("dataset split for '%s' sums to more than 1; will normalize..." % name)
                    if name in self.train_split:
                        self.train_split[name] /= usage
                    if name in self.valid_split:
                        self.valid_split[name] /= usage
                    if name in self.test_split:
                        self.test_split[name] /= usage
        self.skip_verif = thelper.utils.str2bool(config["skip_verif"]) if "skip_verif" in config else True

    @staticmethod
    def _get_collate_fn(val):
        if val is torch.utils.data.dataloader.default_collate:
            return val
        if isinstance(val, dict):
            if "type" not in val or "params" not in val:
                raise AssertionError("unexpected collate function parameter binding dictionary content")
            if not isinstance(val["type"], str):
                raise AssertionError("unexpected collate function type")
            if not isinstance(val["params"], dict):
                raise AssertionError("unexpected collate function params dict type")
            return thelper.utils.import_function(val["type"], val["params"])
        elif isinstance(val, str):
            return thelper.utils.import_function(val)
        else:
            raise AssertionError("unexpected collate val type")

    @staticmethod
    def _get_seed(prefixes, config, stype):
        key = None
        for prefix in prefixes:
            if prefix in config:
                key = prefix
                break
        if key is not None:
            if not isinstance(config[key], stype):
                raise AssertionError("unexpected value type for field '%s'" % key)
            return config[key]
        seed = np.random.randint(2 ** 16)
        logger.info("setting '%s' to %d" % (prefixes[0], seed))
        return seed

    @staticmethod
    def _get_ratios_split(prefix, config):
        key = prefix + "_split"
        if key not in config or not config[key]:
            return {}
        split = config[key]
        if any(ratio < 0 or ratio > 1 for ratio in split.values()):
            raise AssertionError("split ratios in '%s' must be in [0,1]" % key)
        return split

    @staticmethod
    def _get_augments(targets, name, config):
        logger.debug("loading %s augments..." % name)
        for target in targets:
            if target in config and config[target]:
                augments, augments_append = thelper.transforms.load_augments(config[target])
                if augments:
                    logger.debug("will %s %s augments: %s" % ("append" if augments_append else "prefix", name, str(augments)))
                return augments, augments_append
        return None, False

    def _get_raw_split(self, indices):
        for name in self.total_usage:
            if name not in indices:
                raise AssertionError("dataset '%s' does not exist" % name)
        _indices, train_idxs, valid_idxs, test_idxs = {}, {}, {}, {}
        for name, indices in indices.items():
            _indices[name] = copy.deepcopy(indices)
            train_idxs[name] = []
            valid_idxs[name] = []
            test_idxs[name] = []
        indices = _indices
        shuffle = any([self.train_shuffle, self.valid_shuffle, self.test_shuffle])
        if shuffle:
            np.random.seed(self.seeds["test"])  # test idxs will be picked first, then valid+train
            for idxs in indices.values():
                np.random.shuffle(idxs)
        offsets = dict.fromkeys(self.total_usage, 0)
        for loader_idx, (idxs_map, ratio_map) in enumerate(zip([test_idxs, valid_idxs, train_idxs],
                                                               [self.test_split, self.valid_split, self.train_split])):
            for name in self.total_usage.keys():
                if name in ratio_map:
                    count = int(round(ratio_map[name] * len(indices[name])))
                    if count < 0:
                        raise AssertionError("ratios should be non-negative values!")
                    elif count < 1 and len(indices[name]) > 0:
                        logger.warning("split ratio for '%s' too small, sample set will be empty" % name)
                    begidx = offsets[name]
                    endidx = min(begidx + count, len(indices[name]))
                    idxs_map[name] = indices[name][begidx:endidx]
                    offsets[name] = endidx
            if loader_idx == 0 and shuffle:
                np.random.seed(self.seeds["valid"])  # all test idxs are now picked, reshuffle for train/valid
                for name in self.total_usage.keys():
                    trainvalid_idxs = indices[name][offsets[name]:]
                    np.random.shuffle(trainvalid_idxs)
                    indices[name][offsets[name]:] = trainvalid_idxs
        if shuffle:
            np.random.seed(self.seeds["numpy"])  # back to default random state for future use
        return train_idxs, valid_idxs, test_idxs

    def get_split(self, datasets, task):
        r"""Returns the train/valid/test sample indices split for a given dataset (name-parser) map.

        Note that the returned indices are unique, possibly shuffled, and never duplicated between sets.
        If the samples have a class attribute (i.e. the task is related to classification), the split
        will respect the initial distribution and apply the ratios within the classes themselves. For
        example, consider a dataset of three classes (:math:`A`, :math:`B`, and :math:`C`) that contains
        100 samples such as:

        .. math::

            |A| = 50,\;|B| = 30,\;|C| = 20

        If we require a 80%-10%-10% ratio distribution for the training, validation, and test loaders
        respectively, the resulting split will contain the following sample counts:

        .. math::

                \text{training loader} = {40A + 24B + 16C}

        .. math::

                \text{validation loader} = {5A + 3B + 2C}

        .. math::

                \text{test loader} = {5A + 3B + 2C}

        Args:
            datasets: the map of datasets to split, where each has a name (key) and a parser (value).
            task: a task object that should be compatible with all provided datasets (can be ``None``).

        Returns:
            A three-element tuple containing the maps of the training, validation, and test sets
            respectively. These maps associate dataset names to a list of sample indices.
        """
        dataset_sizes = {}
        must_split = {}
        global_size = 0
        for dataset_name, dataset in datasets.items():
            if not isinstance(dataset, thelper.data.Dataset) and not isinstance(dataset, thelper.data.ExternalDataset):
                raise AssertionError("unexpected dataset type for '%s'" % dataset_name)
            dataset_sizes[dataset_name] = len(dataset)
            global_size += dataset_sizes[dataset_name]
            # if a single dataset is used in more than a single loader, we cannot skip the rebalancing below
            must_split[dataset_name] = sum([dataset_name in split for split in
                                            [self.train_split, self.valid_split, self.test_split]]) > 1
        global_size = sum(len(dataset) for dataset in datasets.values())
        logger.info("splitting datasets with parsed sizes = %s" % str(dataset_sizes))
        must_split = any(must_split.values())
        if task is not None and isinstance(task, thelper.tasks.Classification) and not self.skip_class_balancing and must_split:
            # note: with current impl, all class sets will be shuffled the same way... (shouldnt matter, right?)
            logger.debug("will split evenly over %d classes..." % len(task.get_class_names()))
            unset_class_key = "<unset>"
            global_class_names = task.get_class_names() + [unset_class_key]  # extra name added for unlabeled samples (if needed!)
            sample_maps = {}
            for dataset_name, dataset in datasets.items():
                if isinstance(dataset, thelper.data.ExternalDataset):
                    if hasattr(dataset.samples, "samples") and isinstance(dataset.samples.samples, list):
                        sample_maps[dataset_name] = task.get_class_sample_map(dataset.samples.samples, unset_class_key)
                    else:
                        logger.warning(("must fully parse the external dataset '%s' for intra-class shuffling;" % dataset_name) +
                                       " this might take a while!\n(consider making a dataset interface that can return labels" +
                                       " only, it would greatly speed up the analysis of class distributions)")
                        label_key = task.get_gt_key()
                        # to allow glitch-less tqdm printing after latest logger output
                        sys.stdout.flush()
                        sys.stderr.flush()
                        time.sleep(0.01)
                        samples = []
                        for sample in tqdm.tqdm(dataset):
                            if label_key not in sample:
                                raise AssertionError("could not find label key ('%s') in sample dict" % label_key)
                            samples.append({label_key: sample[label_key]})
                        sample_maps[dataset_name] = task.get_class_sample_map(samples, unset_class_key)
                elif isinstance(dataset, thelper.data.Dataset):
                    sample_maps[dataset_name] = task.get_class_sample_map(dataset.samples, unset_class_key)
            train_idxs, valid_idxs, test_idxs = {}, {}, {}
            for class_name in global_class_names:
                curr_class_samples, curr_class_size = {}, {}
                for dataset_name in datasets:
                    class_samples = sample_maps[dataset_name][class_name] if class_name in sample_maps[dataset_name] else []
                    samples_pairs = list(zip(class_samples, [class_name] * len(class_samples)))
                    curr_class_samples[dataset_name] = samples_pairs
                    curr_class_size[dataset_name] = len(curr_class_samples[dataset_name])
                    logger.debug("dataset '{}' class '{}' sample count: {} ({}% of local, {}% of total)".format(
                        dataset_name,
                        class_name,
                        curr_class_size[dataset_name],
                        int(100 * curr_class_size[dataset_name] / dataset_sizes[dataset_name]),
                        int(100 * curr_class_size[dataset_name] / global_size)))
                class_train_idxs, class_valid_idxs, class_test_idxs = self._get_raw_split(curr_class_samples)
                for idxs_dict_list, class_idxs_dict_list in zip([train_idxs, valid_idxs, test_idxs],
                                                                [class_train_idxs, class_valid_idxs, class_test_idxs]):
                    for dataset_name in datasets:
                        if dataset_name in idxs_dict_list:
                            idxs_dict_list[dataset_name] += class_idxs_dict_list[dataset_name]
                        else:
                            idxs_dict_list[dataset_name] = class_idxs_dict_list[dataset_name]
        else:  # no balancing to be done
            dataset_indices = {}
            for dataset_name in datasets:
                # note: all indices paired with 'None' below as class is ignored; used for compatibility with code above
                dataset_indices[dataset_name] = list(
                    zip(list(range(dataset_sizes[dataset_name])), [None] * dataset_sizes[dataset_name]))
            train_idxs, valid_idxs, test_idxs = self._get_raw_split(dataset_indices)
        return train_idxs, valid_idxs, test_idxs

    def create_loaders(self, datasets, train_idxs, valid_idxs, test_idxs):
        """Returns the data loaders for the train/valid/test sets based on a prior split.

        This function essentially takes the dataset parser interfaces and indices maps, and instantiates
        data loaders that are ready to produce samples for training or evaluation. Note that the dataset
        parsers will be deep-copied in each data loader, meaning that they should ideally not contain a
        persistent loading state or a large buffer.

        Args:
            datasets: the map of dataset parsers, where each has a name (key) and a parser (value).
            train_idxs: training data samples indices map.
            valid_idxs: validation data samples indices map.
            test_idxs: test data samples indices map.

        Returns:
            A three-element tuple containing the training, validation, and test data loaders, respectively.
        """
        loaders = []
        for idxs_map, (augs, augs_append), shuffle, scale, sampler_apply, batch_size, collate_fn \
                in zip([train_idxs, valid_idxs, test_idxs],
                       [(self.train_augments, self.train_augments_append),
                        (self.valid_augments, self.valid_augments_append),
                        (self.test_augments, self.test_augments_append)],
                       [self.train_shuffle, self.valid_shuffle, self.test_shuffle],
                       [self.train_scale, self.valid_scale, self.test_scale],
                       [self.train_sampler, self.valid_sampler, self.test_sampler],
                       [self.train_batch_size, self.valid_batch_size, self.test_batch_size],
                       [self.train_collate_fn, self.valid_collate_fn, self.test_collate_fn]):
            loader_sample_idx_offset = 0
            loader_sample_classes = []
            loader_sample_idxs = []
            loader_datasets = []
            for dataset_name, sample_idxs in idxs_map.items():
                if not sample_idxs:
                    continue
                # todo: investigate need to copy at all?
                if not datasets[dataset_name].deepcopy:
                    dataset = copy.copy(datasets[dataset_name])
                else:
                    dataset = copy.deepcopy(datasets[dataset_name])
                if augs:
                    augs_copy = copy.deepcopy(augs)
                    if dataset.transforms is not None:
                        if augs_append:
                            dataset.transforms = thelper.transforms.Compose([dataset.transforms, augs_copy])
                        else:
                            dataset.transforms = thelper.transforms.Compose([augs_copy, dataset.transforms])
                    else:
                        dataset.transforms = augs_copy
                for sample_idx_idx in range(len(sample_idxs)):
                    # values were paired in tuples earlier, 0=idx, 1=label
                    loader_sample_idxs.append(sample_idxs[sample_idx_idx][0] + loader_sample_idx_offset)
                    loader_sample_classes.append(sample_idxs[sample_idx_idx][1])
                loader_sample_idx_offset += len(dataset)
                loader_datasets.append(dataset)
            if len(loader_datasets) > 0:
                dataset = torch.utils.data.ConcatDataset(loader_datasets) if len(loader_datasets) > 1 else loader_datasets[0]
                if self.sampler_type is not None and sampler_apply:
                    sampler_params = {**self.sampler_params}
                    if self.sampler_pass_labels:
                        sampler_params[self.sampler_pass_labels_param_name] = loader_sample_classes
                    sampler_sig = inspect.signature(self.sampler_type)
                    if "seeds" in sampler_sig.parameters:
                        sampler_params["seeds"] = self.seeds
                    if "scale" in sampler_sig.parameters:
                        sampler_params["scale"] = scale
                    elif scale != 1.0:
                        raise AssertionError("could not apply scale factor to sample with type '%s'" % str(self.sampler_type))
                    sampler = self.sampler_type(loader_sample_idxs, **self.sampler_params)
                else:
                    if shuffle:
                        sampler = thelper.data.SubsetRandomSampler(loader_sample_idxs, seeds=self.seeds, scale=scale)
                    else:
                        if scale != 1.0:
                            raise AssertionError("sequential sampler currently does not handle scale changes")
                        sampler = thelper.data.SubsetSequentialSampler(loader_sample_idxs)
                if not hasattr(sampler, "__len__"):
                    raise AssertionError("sampler should always expose sample count via '__len__'")
                if batch_size is None or batch_size <= 0:
                    raise AssertionError("invalid batch size")
                loaders.append(DataLoader(dataset=dataset, batch_size=batch_size, sampler=sampler,
                                          num_workers=self.workers, collate_fn=collate_fn,
                                          pin_memory=self.pin_memory, drop_last=self.drop_last,
                                          seeds=self.seeds))
            else:
                loaders.append(None)
        train_loader, valid_loader, test_loader = loaders
        logger.info("initialized loaders with batch counts:\n  train=%d\n  valid=%d\n  test=%d" %
                    (len(train_loader) if train_loader else 0,
                     len(valid_loader) if valid_loader else 0,
                     len(test_loader) if test_loader else 0))
        return train_loader, valid_loader, test_loader

    def get_base_transforms(self):
        """Returns the (global) sample transformation operations parsed in the data configuration."""
        return self.base_transforms



# noinspection PyPackageRequirements
import mock
import numpy as np

import thelper
import thelper.transforms


# noinspection PyUnusedLocal
@mock.patch("thelper.transforms.CenterCrop", autospec=True)
@mock.patch("thelper.transforms.RandomResizedCrop", autospec=True)
def test_transform_pipeline_construct(fake_op_class1, fake_op_class2):
    _ = thelper.transforms.load_transforms([  # noqa: F841
        {
            "operation": "thelper.transforms.RandomResizedCrop",
            "params": {
                "output_size": [100, 100]
            }
        },
        {
            "operation": "thelper.transforms.CenterCrop",
            "params": {
                "size": [200, 200]
            }
        }
    ])
    fake_op_class1.assert_called_once_with(output_size=[100, 100])
    fake_op_class2.assert_called_once_with(size=[200, 200])


# noinspection PyUnusedLocal
@mock.patch.object(thelper.transforms.CenterCrop, "__call__")
@mock.patch.object(thelper.transforms.RandomResizedCrop, "__call__")
def test_transform_stochastic_pipeline(fake_op1, fake_op2):
    fake_op1.side_effect = lambda x: x
    fake_op2.side_effect = lambda x: x
    sample = np.random.rand(5, 4, 3)
    transforms = thelper.transforms.load_transforms([
        {
            "operation": "thelper.transforms.RandomResizedCrop",
            "params": {
                "output_size": [100, 100]
            }
        },
        {
            "operation": "thelper.transforms.CenterCrop",
            "params": {
                "size": [200, 200]
            }
        }
    ])
    out = transforms(sample)
    assert np.array_equal(out, sample)



import torch
import torch.nn


class AddCoords(torch.nn.Module):
    def __init__(self, radius_channel=False):
        super().__init__()
        self.radius_channel = radius_channel

    def forward(self, in_tensor):
        """
        in_tensor: (batch_size, channels, x_dim, y_dim)
        [0,0,0,0]   [0,1,2,3]
        [1,1,1,1]   [0,1,2,3]    << (i,j)th coordinates of pixels added as separate channels
        [2,2,2,2]   [0,1,2,3]
        taken from mkocabas.
        """
        batch_size_tensor = in_tensor.shape[0]

        xx_ones = torch.ones([1, in_tensor.shape[2]], dtype=torch.int32)
        xx_ones = xx_ones.unsqueeze(-1)

        xx_range = torch.arange(in_tensor.shape[2], dtype=torch.int32).unsqueeze(0)
        xx_range = xx_range.unsqueeze(1)

        xx_channel = torch.matmul(xx_ones, xx_range)
        xx_channel = xx_channel.unsqueeze(-1)

        yy_ones = torch.ones([1, in_tensor.shape[3]], dtype=torch.int32)
        yy_ones = yy_ones.unsqueeze(1)

        yy_range = torch.arange(in_tensor.shape[3], dtype=torch.int32).unsqueeze(0)
        yy_range = yy_range.unsqueeze(-1)

        yy_channel = torch.matmul(yy_range, yy_ones)
        yy_channel = yy_channel.unsqueeze(-1)

        xx_channel = xx_channel.permute(0, 3, 1, 2)
        yy_channel = yy_channel.permute(0, 3, 1, 2)

        xx_channel = xx_channel.float() / (in_tensor.shape[2] - 1)
        yy_channel = yy_channel.float() / (in_tensor.shape[3] - 1)

        xx_channel = xx_channel * 2 - 1
        yy_channel = yy_channel * 2 - 1

        xx_channel = xx_channel.repeat(batch_size_tensor, 1, 1, 1)
        yy_channel = yy_channel.repeat(batch_size_tensor, 1, 1, 1)

        dev = in_tensor.device
        out = torch.cat([in_tensor, xx_channel.to(dev), yy_channel.to(dev)], dim=1)
        if self.radius_channel:
            radius_calc = torch.sqrt(torch.pow(xx_channel - 0.5, 2) + torch.pow(yy_channel - 0.5, 2))
            out = torch.cat([out, radius_calc.to(dev)], dim=1)
        return out


class CoordConv2d(torch.nn.Module):
    """ add any additional coordinate channels to the input tensor """
    def __init__(self, in_channels, *args, radius_channel=False, **kwargs):
        super().__init__()
        self.addcoord = AddCoords(radius_channel=radius_channel)
        extra_in_channels = 3 if radius_channel else 2
        self.conv = torch.nn.Conv2d(in_channels + extra_in_channels, *args, **kwargs)

    def forward(self, in_tensor):
        out = self.addcoord(in_tensor)
        out = self.conv(out)
        return out


class CoordConvTranspose(torch.nn.Module):
    """CoordConvTranspose layer for segmentation tasks."""
    def __init__(self, radius_channel, *args, **kwargs):
        super().__init__()
        self.addcoord = AddCoords(radius_channel=radius_channel)
        self.convT = torch.nn.ConvTranspose2d(*args, **kwargs)

    def forward(self, in_tensor):
        out = self.addcoord(in_tensor)
        out = self.convT(out)
        return out



import re
from collections import OrderedDict

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.utils.model_zoo as model_zoo

__all__ = ['DenseNet', 'densenet121', 'densenet169', 'densenet201', 'densenet161']

model_urls = {
    'densenet121': 'https://download.pytorch.org/models/densenet121-a639ec97.pth',
    'densenet169': 'https://download.pytorch.org/models/densenet169-b2777c0a.pth',
    'densenet201': 'https://download.pytorch.org/models/densenet201-c1103571.pth',
    'densenet161': 'https://download.pytorch.org/models/densenet161-8d451a50.pth',
}


def densenet121(pretrained=False, **kwargs):
    """Densenet-121 model from `"Densely Connected Convolutional Networks" <https://arxiv.org/pdf/1608.06993.pdf>`_

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = DenseNet(num_init_features=64, growth_rate=32, block_config=(6, 12, 24, 16),
                     **kwargs)
    if pretrained:
        # '.'s are no longer allowed in module names, but pervious _DenseLayer
        # has keys 'norm.1', 'relu.1', 'conv.1', 'norm.2', 'relu.2', 'conv.2'.
        # They are also in the checkpoints in model_urls. This pattern is used
        # to find such keys.
        pattern = re.compile(
            r'^(.*denselayer\d+\.(?:norm|relu|conv))\.((?:[12])\.(?:weight|bias|running_mean|running_var))$')
        state_dict = model_zoo.load_url(model_urls['densenet121'])
        for key in list(state_dict.keys()):
            res = pattern.match(key)
            if res:
                new_key = res.group(1) + res.group(2)
                state_dict[new_key] = state_dict[key]
                del state_dict[key]
        model.load_state_dict(state_dict)
    return model


def densenet169(pretrained=False, **kwargs):
    """Densenet-169 model from `"Densely Connected Convolutional Networks" <https://arxiv.org/pdf/1608.06993.pdf>`_

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = DenseNet(num_init_features=64, growth_rate=32, block_config=(6, 12, 32, 32),
                     **kwargs)
    if pretrained:
        # '.'s are no longer allowed in module names, but pervious _DenseLayer
        # has keys 'norm.1', 'relu.1', 'conv.1', 'norm.2', 'relu.2', 'conv.2'.
        # They are also in the checkpoints in model_urls. This pattern is used
        # to find such keys.
        pattern = re.compile(
            r'^(.*denselayer\d+\.(?:norm|relu|conv))\.((?:[12])\.(?:weight|bias|running_mean|running_var))$')
        state_dict = model_zoo.load_url(model_urls['densenet169'])
        for key in list(state_dict.keys()):
            res = pattern.match(key)
            if res:
                new_key = res.group(1) + res.group(2)
                state_dict[new_key] = state_dict[key]
                del state_dict[key]
        model.load_state_dict(state_dict)
    return model


def densenet201(pretrained=False, **kwargs):
    """Densenet-201 model from `"Densely Connected Convolutional Networks" <https://arxiv.org/pdf/1608.06993.pdf>`_

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = DenseNet(num_init_features=64, growth_rate=32, block_config=(6, 12, 48, 32),
                     **kwargs)
    if pretrained:
        # '.'s are no longer allowed in module names, but pervious _DenseLayer
        # has keys 'norm.1', 'relu.1', 'conv.1', 'norm.2', 'relu.2', 'conv.2'.
        # They are also in the checkpoints in model_urls. This pattern is used
        # to find such keys.
        pattern = re.compile(
            r'^(.*denselayer\d+\.(?:norm|relu|conv))\.((?:[12])\.(?:weight|bias|running_mean|running_var))$')
        state_dict = model_zoo.load_url(model_urls['densenet201'])
        for key in list(state_dict.keys()):
            res = pattern.match(key)
            if res:
                new_key = res.group(1) + res.group(2)
                state_dict[new_key] = state_dict[key]
                del state_dict[key]
        model.load_state_dict(state_dict)
    return model


def densenet161(pretrained=False, **kwargs):
    """Densenet-161 model from `"Densely Connected Convolutional Networks" <https://arxiv.org/pdf/1608.06993.pdf>`_

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = DenseNet(num_init_features=96, growth_rate=48, block_config=(6, 12, 36, 24),
                     **kwargs)
    if pretrained:
        # '.'s are no longer allowed in module names, but pervious _DenseLayer
        # has keys 'norm.1', 'relu.1', 'conv.1', 'norm.2', 'relu.2', 'conv.2'.
        # They are also in the checkpoints in model_urls. This pattern is used
        # to find such keys.
        pattern = re.compile(
            r'^(.*denselayer\d+\.(?:norm|relu|conv))\.((?:[12])\.(?:weight|bias|running_mean|running_var))$')
        state_dict = model_zoo.load_url(model_urls['densenet161'])
        for key in list(state_dict.keys()):
            res = pattern.match(key)
            if res:
                new_key = res.group(1) + res.group(2)
                state_dict[new_key] = state_dict[key]
                del state_dict[key]
        model.load_state_dict(state_dict)
    return model


class SqueezeExcitationLayer(torch.nn.Module):

    def __init__(self, channel, reduction=16):
        super().__init__()
        self.pool = torch.nn.AdaptiveAvgPool2d(1)
        self.fc = torch.nn.Sequential(
            torch.nn.Linear(channel, channel // reduction),
            torch.nn.ReLU(inplace=True),
            torch.nn.Linear(channel // reduction, channel),
            torch.nn.Sigmoid()
        )

    def forward(self, x):
        b, c, _, _ = x.size()
        y = self.pool(x).view(b, c)
        y = self.fc(y).view(b, c, 1, 1)
        return x * y


class _DenseLayer(nn.Sequential):
    def __init__(self, num_input_features, growth_rate, bn_size, drop_rate):
        super().__init__()
        self.add_module('norm1', nn.BatchNorm2d(num_input_features)),
        self.add_module('relu1', nn.ReLU(inplace=True)),
        self.add_module('conv1', nn.Conv2d(num_input_features, bn_size *
                                           growth_rate, kernel_size=1, stride=1, bias=False)),
        self.add_module('norm2', nn.BatchNorm2d(bn_size * growth_rate)),
        self.add_module('relu2', nn.ReLU(inplace=True)),
        self.add_module('conv2', nn.Conv2d(bn_size * growth_rate, growth_rate,
                                           kernel_size=3, stride=1, padding=1, bias=False)),

        self.drop_rate = drop_rate

    def forward(self, x):
        new_features = super().forward(x)
        if self.drop_rate > 0:
            new_features = F.dropout(new_features, p=self.drop_rate, training=self.training)
        return torch.cat([x, new_features], 1)


class _DenseBlock(nn.Sequential):
    def __init__(self, num_layers, num_input_features, bn_size, growth_rate, drop_rate):
        super().__init__()
        for i in range(num_layers):
            layer = _DenseLayer(num_input_features + i * growth_rate, growth_rate, bn_size, drop_rate)
            self.add_module('denselayer%d' % (i + 1), layer)


class _DenseSeLayer(nn.Sequential):
    def __init__(self, num_input_features, growth_rate, bn_size, drop_rate):
        super().__init__()
        self.add_module('norm1', nn.BatchNorm2d(num_input_features)),
        self.add_module('relu1', nn.ReLU(inplace=True)),
        self.add_module('conv1', nn.Conv2d(num_input_features, bn_size *
                                           growth_rate, kernel_size=1, stride=1, bias=False)),
        self.add_module('norm2', nn.BatchNorm2d(bn_size * growth_rate)),
        self.add_module('relu2', nn.ReLU(inplace=True)),
        self.add_module('conv2', nn.Conv2d(bn_size * growth_rate, growth_rate,
                                           kernel_size=3, stride=1, padding=1, bias=False)),
        self.add_module('se', SqueezeExcitationLayer(growth_rate)),
        self.drop_rate = drop_rate

    def forward(self, x):
        new_features = super().forward(x)
        if self.drop_rate > 0:
            new_features = F.dropout(new_features, p=self.drop_rate, training=self.training)
        return torch.cat([x, new_features], 1)


class _DenseSeBlock(nn.Sequential):
    def __init__(self, num_layers, num_input_features, bn_size, growth_rate, drop_rate):
        super().__init__()
        for i in range(num_layers):
            layer = _DenseSeLayer(num_input_features + i * growth_rate, growth_rate, bn_size, drop_rate)
            self.add_module('denselayer%d' % (i + 1), layer)


class _Transition(nn.Sequential):
    def __init__(self, num_input_features, num_output_features):
        super().__init__()
        self.add_module('norm', nn.BatchNorm2d(num_input_features))
        self.add_module('relu', nn.ReLU(inplace=True))
        self.add_module('conv', nn.Conv2d(num_input_features, num_output_features,
                                          kernel_size=1, stride=1, bias=False))
        self.add_module('pool', nn.AvgPool2d(kernel_size=2, stride=2))


class DenseNet(nn.Module):
    """Densenet-BC model class, based on `"Densely Connected Convolutional Networks" <https://arxiv.org/pdf/1608.06993.pdf>`_

    Args:
        growth_rate (int): how many filters to add each layer (`k` in paper)
        block_config (list of 4 ints): how many layers in each pooling block
        num_init_features (int): the number of filters to learn in the first convolution layer
        bn_size (int): multiplicative factor for number of bottle neck layers (i.e. `bn_size * k features`
            in the bottleneck layer)
        drop_rate (float): dropout rate after each dense layer
        num_classes (int): number of classification classes
    """

    def __init__(self, growth_rate=32, block_config=(6, 12, 24, 16),
                 num_init_features=64, bn_size=4, drop_rate=0, num_classes=1000):
        super().__init__()
        # First convolution
        self.features = nn.Sequential(OrderedDict([
            ('conv0', nn.Conv2d(3, num_init_features, kernel_size=7, stride=2, padding=3, bias=False)),
            ('norm0', nn.BatchNorm2d(num_init_features)),
            ('relu0', nn.ReLU(inplace=True)),
            ('pool0', nn.MaxPool2d(kernel_size=3, stride=2, padding=1)),
        ]))
        # Each denseblock
        num_features = num_init_features
        for i, num_layers in enumerate(block_config):
            block = _DenseBlock(num_layers=num_layers, num_input_features=num_features,
                                bn_size=bn_size, growth_rate=growth_rate, drop_rate=drop_rate)
            self.features.add_module('denseblock%d' % (i + 1), block)
            num_features = num_features + num_layers * growth_rate
            if i != len(block_config) - 1:
                trans = _Transition(num_input_features=num_features, num_output_features=num_features // 2)
                self.features.add_module('transition%d' % (i + 1), trans)
                num_features = num_features // 2
        # Final batch norm
        self.features.add_module('norm5', nn.BatchNorm2d(num_features))
        # Linear layer
        self.classifier = nn.Linear(num_features, num_classes)
        # Official init from torch repo.
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.Linear):
                nn.init.constant_(m.bias, 0)

    def forward(self, x):
        features = self.features(x)
        out = F.relu(features, inplace=True)
        out = F.avg_pool2d(out, kernel_size=7, stride=1).view(features.size(0), -1)
        out = self.classifier(out)
        return out



"""PASCAL VOC dataset parser module.

This module contains a dataset parser used to load the PASCAL Visual Object Classes (VOC) dataset for
semantic segmentation or object detection. See http://host.robots.ox.ac.uk/pascal/VOC/ for more info.
"""

import logging
import os
import xml.etree.ElementTree

import cv2 as cv
import numpy as np

import thelper.tasks
import thelper.utils
from thelper.data.parsers import Dataset

logger = logging.getLogger(__name__)


class PASCALVOC(Dataset):
    """PASCAL VOC dataset parser.

    This class can be used to parse the PASCAL VOC dataset for either semantic segmentation or object
    detection. The task object it exposes will be changed accordingly. In all cases, the 2012 version
    of the dataset will be used.

    TODO: Finish implementation of object detection task setup.
    TODO: Add support for semantic instance segmentation.

    .. seealso::
        | :class:`thelper.data.parsers.Dataset`
    """

    _label_name_map = {
        "background": 0,
        "aeroplane": 1,
        "bicycle": 2,
        "bird": 3,
        "boat": 4,
        "bottle": 5,
        "bus": 6,
        "car": 7,
        "cat": 8,
        "chair": 9,
        "cow": 10,
        "diningtable": 11,
        "dog": 12,
        "horse": 13,
        "motorbike": 14,
        "person": 15,
        "pottedplant": 16,
        "sheep": 17,
        "sofa": 18,
        "train": 19,
        "tvmonitor": 20,
        "dontcare": 255,
    }

    _dontcare_val = 255
    _supported_tasks = ["detect", "segm"]
    _supported_subsets = ["train", "trainval", "val", "test"]
    _train_archive_name = "VOCtrainval_11-May-2012.tar"
    _test_archive_name = "VOC2012test.tar"
    _train_archive_url = "http://pjreddie.com/media/files/" + _train_archive_name
    _test_archive_url = "http://pjreddie.com/media/files/" + _test_archive_name
    _train_archive_md5 = "6cd6e144f989b92b3379bac3b3de84fd"
    _test_archive_md5 = "9065beb292b6c291fad82b2725749fda"

    def __init__(self, config, transforms=None):
        self.task_name = str(thelper.utils.get_key_def("task", config, "segm")).lower()
        if self.task_name not in self._supported_tasks:
            raise AssertionError("unrecognized task type '%s'" % self.task_name)
        subset = str(thelper.utils.get_key_def("subset", config, "trainval")).lower()
        if subset not in self._supported_subsets:
            raise AssertionError("unrecognized data subset '%s'" % subset)
        download = thelper.utils.str2bool(thelper.utils.get_key_def("download", config, False))
        root = os.path.abspath(thelper.utils.get_key("root", config))
        devkit_path = os.path.join(root, "VOCdevkit")
        if not os.path.isdir(devkit_path):
            if not download:
                raise AssertionError("invalid devkit path '%s'" % devkit_path)
            else:
                logger.info("downloading training data archive...")
                train_archive_path = thelper.utils.download_file(self._train_archive_url,
                                                                 root, self._train_archive_name,
                                                                 self._train_archive_md5)
                logger.info("extracting training data archive...")
                thelper.utils.extract_tar(train_archive_path, root, flags="r:")
                logger.info("downloading test data archive...")
                test_archive_path = thelper.utils.download_file(self._test_archive_url,
                                                                root, self._test_archive_name,
                                                                self._test_archive_md5)
                logger.info("extracting test data archive...")
                thelper.utils.extract_tar(test_archive_path, root, flags="r:")
        if not os.path.isdir(devkit_path):
            raise AssertionError("messed up tar extraction")
        dataset_path = os.path.join(devkit_path, "VOC2012")
        if not os.path.isdir(dataset_path):
            raise AssertionError("could not locate dataset folder 'VOC2012' at '%s'" % devkit_path)
        imagesets_path = os.path.join(dataset_path, "ImageSets")
        if not os.path.isdir(dataset_path):
            raise AssertionError("could not locate image sets folder at '%s'" % imagesets_path)
        super().__init__(config=config, transforms=transforms)
        self.preload = thelper.utils.str2bool(thelper.utils.get_key_def("preload", config, True))  # @@@ CHANGE TODO
        # should use_difficult be true for training, but false for validation?
        use_difficult = thelper.utils.str2bool(thelper.utils.get_key_def("use_difficult", config, False))
        use_occluded = thelper.utils.str2bool(thelper.utils.get_key_def("use_occluded", config, True))
        use_truncated = thelper.utils.str2bool(thelper.utils.get_key_def("use_truncated", config, True))
        self.image_key = thelper.utils.get_key_def("image_key", config, "image")
        self.sample_name_key = thelper.utils.get_key_def("name_key", config, "name")
        self.image_path_key = thelper.utils.get_key_def("image_path_key", config, "image_path")
        self.gt_path_key = thelper.utils.get_key_def("gt_path_key", config, "gt_path")
        meta_keys = [self.sample_name_key, self.image_path_key, self.gt_path_key]
        self.gt_key = None
        self.task = None
        image_set_name = None
        valid_sample_names = None
        target_labels = thelper.utils.get_key_def("target_labels", config, None)
        if target_labels is not None:
            if not isinstance(target_labels, list):
                target_labels = [target_labels]
            if not target_labels or not all([label in self._label_name_map for label in target_labels]):
                raise AssertionError("target labels should be given as list of names (strings) that already exist")
            self.label_name_map = {}
            for name in self._label_name_map:
                if name in target_labels or name == "background" or name == "dontcare":
                    self.label_name_map[name] = len(self.label_name_map) if name != "dontcare" else self._dontcare_val
        else:
            self.label_name_map = self._label_name_map
        self.label_colors = {idx: thelper.utils.get_label_color_mapping(self._label_name_map[name])[::-1]
                             for name, idx in self.label_name_map.items()}
        if self.task_name == "detect":
            self.gt_key = thelper.utils.get_key_def("bboxes_key", config, "bboxes")
            # self.task = thelper.tasks.Detection(...)
            image_set_name = "Main"
            raise NotImplementedError
        elif self.task_name == "segm":
            self.gt_key = thelper.utils.get_key_def("label_map_key", config, "label_map")
            color_map = {name: self.label_colors[idx][::-1] for name, idx in self.label_name_map.items()}
            self.task = thelper.tasks.Segmentation(self.label_name_map, input_key=self.image_key,
                                                   label_map_key=self.gt_key, meta_keys=meta_keys,
                                                   dontcare=self._dontcare_val,
                                                   color_map=color_map)
            image_set_name = "Segmentation"
            # if using target labels, must rely on image set luts to confirm content
            if target_labels is not None:
                valid_sample_names = set()
                for label in self.label_name_map:
                    if label == "background" or label == "dontcare":
                        continue
                    with open(os.path.join(imagesets_path, "Main", label + "_" + subset + ".txt")) as image_subset_fd:
                        for line in image_subset_fd:
                            sample_name, val = line.split()
                            if int(val) > 0:
                                valid_sample_names.add(sample_name)
        imageset_path = os.path.join(imagesets_path, image_set_name, subset + ".txt")
        if not os.path.isfile(imageset_path):
            raise AssertionError("cannot locate sample set file at '%s'" % imageset_path)
        image_folder_path = os.path.join(dataset_path, "JPEGImages")
        if not os.path.isdir(image_folder_path):
            raise AssertionError("cannot locate image folder at '%s'" % image_folder_path)
        with open(imageset_path) as image_subset_fd:
            if valid_sample_names is None:
                sample_names = image_subset_fd.read().splitlines()
            else:
                sample_names = set()
                for sample_name in image_subset_fd:
                    sample_name = sample_name.strip()
                    if sample_name in valid_sample_names:
                        sample_names.add(sample_name)
                sample_names = list(sample_names)
        action = "preloading" if self.preload else "initializing"
        logger.info("%s pascal voc dataset for task='%s' and set='%s'..." % (action, self.task_name, subset))
        self.samples = []
        if self.preload:
            from tqdm import tqdm
        else:
            def tqdm(x):
                return x
        for sample_name in tqdm(sample_names):
            annotation_file_path = os.path.join(dataset_path, "Annotations", sample_name + ".xml")
            if not os.path.isfile(annotation_file_path):
                raise AssertionError("cannot load annotation file for sample '%s'" % sample_name)
            annotation = xml.etree.ElementTree.parse(annotation_file_path).getroot()
            if annotation.tag != "annotation":
                raise AssertionError("unexpected xml content")
            image_path = os.path.join(image_folder_path, annotation.find("filename").text)
            if not os.path.isfile(image_path):
                raise AssertionError("cannot locate image for sample '%s'" % sample_name)
            image = None
            if self.preload:
                image = cv.imread(image_path)
                if image is None:
                    raise AssertionError("could not load image '%s' via opencv" % image_path)
            gt, gt_path = None, None
            if self.task_name == "segm":
                if int(annotation.find("segmented").text) != 1:
                    raise AssertionError("unexpected segmented flag for sample '%s'" % sample_name)
                gt_path = os.path.join(dataset_path, "SegmentationClass", sample_name + ".png")
                if self.preload:
                    gt = cv.imread(gt_path)
                    if gt is None or gt.shape != image.shape:
                        raise AssertionError("unexpected gt shape for sample '%s'" % sample_name)
                    gt = self.encode_label_map(gt)
                    #gt_decoded = self.decode_label_map(gt)
                    #if not np.array_equal(cv.imread(gt_path), gt_decoded):
                    #    raise AssertionError("messed up encoding/decoding functions")
            elif self.task_name == "detect":
                gt_path = annotation_file_path
                gt = []
                for obj in annotation.iter("object"):
                    # TODO: update w/ task-compat structure
                    if not use_difficult and obj.find("difficult").text == "1":
                        continue
                    if not use_occluded and obj.find("occluded").text == "1":
                        continue
                    if not use_truncated and obj.find("truncated").text == "1":
                        continue
                    bbox = obj.find("bndbox")
                    label = obj.find("name").text
                    if label not in self.label_name_map:
                        continue  # user is skipping some labels from the complete set
                    gt.append({
                        "label": obj.find("name").text,
                        "id": self.label_name_map[label],
                        "bbox": {
                            "xmax": bbox.find("xmax").text,
                            "xmin": bbox.find("xmin").text,
                            "ymax": bbox.find("ymax").text,
                            "ymin": bbox.find("ymin").text,
                        },
                        "difficult": thelper.utils.str2bool(obj.find("difficult").text),
                        "occluded": thelper.utils.str2bool(obj.find("occluded").text),
                        "truncated": thelper.utils.str2bool(obj.find("truncated").text),
                    })
                if not gt:
                    continue
            self.samples.append({
                self.sample_name_key: sample_name,
                self.image_path_key: image_path,
                self.gt_path_key: gt_path,
                self.image_key: image,
                self.gt_key: gt,
            })
        logger.info("initialized %d samples" % len(self.samples))

    def __getitem__(self, idx):
        """Returns the data sample (a dictionary) for a specific (0-based) index."""
        sample = self.samples[idx]
        if not self.preload:
            image = cv.imread(sample[self.image_path_key])
            if image is None:
                raise AssertionError("could not load image '%s' via opencv" % sample[self.image_path_key])
            image = image[..., ::-1]  # BGR to RGB
            gt = None
            if self.task_name == "segm":
                gt = cv.imread(sample[self.gt_path_key])
                if gt is None or gt.shape != image.shape:
                    raise AssertionError("unexpected gt shape for sample '%s'" % sample[self.sample_name_key])
                gt = self.encode_label_map(gt)
        else:
            image = sample[self.image_key]
            gt = sample[self.gt_key]
        sample = {
            self.sample_name_key: sample[self.sample_name_key],
            self.image_path_key: sample[self.image_path_key],
            self.gt_path_key: sample[self.gt_path_key],
            self.image_key: image,
            self.gt_key: gt,
        }
        if self.transforms:
            sample = self.transforms(sample)
        if isinstance(sample[self.image_key], np.ndarray) and any([s < 0 for s in sample[self.image_key].strides]):
            # fix unsupported negative strides in PyTorch <= 0.4.0
            sample[self.image_key] = sample[self.image_key].copy()
        return sample

    def get_task(self):
        """Returns the dataset task object that provides the i/o keys for parsing sample dicts."""
        return self.task

    def decode_label_map(self, label_map):
        """Returns a color image from a label indices map."""
        if not isinstance(label_map, np.ndarray) or label_map.ndim != 2:
            raise AssertionError("unexpected label map type/shape, should be 2D np.ndarray")
        output = np.full(list(label_map.shape) + [3], fill_value=self._dontcare_val, dtype=np.uint8)
        for label_idx, label_color in self.label_colors.items():
            output[np.where(label_map == label_idx)] = label_color
        return output

    def encode_label_map(self, label_map):
        """Returns a map of label indices from a color image."""
        if not isinstance(label_map, np.ndarray) or label_map.ndim != 3 or label_map.dtype != np.uint8:
            raise AssertionError("unexpected label map type/shape, should be 3D np.ndarray")
        output = np.full(label_map.shape[:2], fill_value=self._dontcare_val, dtype=np.uint8)
        # TODO: loss might not like uint8, check for useless convs later
        for label_idx, label_color in self.label_colors.items():
            output = np.where(np.all(label_map == label_color, axis=2), label_idx, output)
        return output



"""Model training/evaluation base interface module.

This module contains the interface required to train and/or evaluate a model based on different tasks. The trainers
based on this interface are instantiated in launched sessions based on configuration dictionaries.
"""
import functools
import json
import logging
import math
import os
import platform
import random
import time
from abc import abstractmethod
from copy import deepcopy
from typing import AnyStr  # noqa: F401

import cv2 as cv
import numpy as np
import torch
import torch.optim

import thelper.typedefs as typ  # noqa: F401
import thelper.utils

logger = logging.getLogger(__name__)


class Trainer:
    """Abstract trainer interface that defines basic session i/o and setup operations.

    This interface defines the general behavior of a training session which includes configuration parsing, tensorboard
    setup, metrics and goal setup, and loss/optimizer setup. It also provides utilities for uploading models and tensors
    on specific devices, and for saving the state of a session. This interface should be specialized for every task by
    implementing the ``_train_epoch`` and ``_eval_epoch`` functions in a derived class. See
    :class:`thelper.train.classif.ImageClassifTrainer` for an example.

    The main parameters that will be parsed by this interface from a configuration dictionary are the following:

    - ``epochs`` (mandatory if training): number of epochs to train for; one epoch is one iteration over all mini-batches.
    - ``optimization`` (mandatory if training): sub-dictionary containing types and extra parameters required for
      instantiating the loss, optimizer, and scheduler objects. See the code of each related loading function for more
      information on special parameters.
    - ``save_freq`` (optional, default=1): checkpoint save frequency (will save every epoch multiple of given number).
    - ``save_raw`` (optional, default=True): specifies whether to save raw types or thelper objects in checkpoints.
    - ``use_tbx`` (optional, default=False): defines whether to use tensorboardX writers for logging or not.
    - ``device`` (optional): specifies which device to train/evaluate the model on (default=all available).
    - ``metrics``: list of metrics to instantiate and update during training/evaluation; see related loading function for
      more information.
    - ``monitor``: specifies the name of the metric that should be monitored on the validation set for model improvement.

    Example configuration file::

        # ...
        "trainer": {
            # type of trainer to instantiate (linked to task type)
            "type": "thelper.train.ImageClassifTrainer",
            # train for 40 epochs
            "epochs": 40,
            # save every 5 epochs
            "save_freq": 5,
            # monitor validation accuracy and save best model based on that
            "monitor": "accuracy",
            # optimization parameters block
            "optimization": {
                # all types & params below provided by PyTorch
                "loss": {
                    "type": "torch.nn.CrossEntropyLoss"
                },
                "optimizer": {
                    "type": "torch.optim.SGD",
                    "params": {
                        "lr": 0.1,
                        "momentum": 0.9,
                        "weight_decay": 1e-06,
                        "nesterov": true
                    }
                },
                "scheduler": {
                    "type": "torch.optim.lr_scheduler.StepLR",
                    "params": {
                        "step_size": 10,
                        "step_size": 0.1
                    }
                }
            },
            # in this example, we use two metrics in total
            # (one for monitoring, and one for logging only)
            "metrics": {
                "accuracy": {
                    "type": "thelper.optim.CategoryAccuracy"
                },
                "fullreport": {
                    "type": "thelper.optim.ClassifReport"
                }
            }
        }
        # ...

    Attributes:
        logger: used to output debug/warning/error messages to session log.
        name: name of the session, used for printing and creating log folders.
        epochs: number of epochs to train the model for.
        optimization_config: dictionary of optim-related parameters, parsed at training time.
        save_freq: frequency of checkpoint saves while training (i.e. save every X epochs).
        save_raw: specifies whether to save raw types or thelper objects in checkpoints.
        checkpoint_dir: session checkpoint output directory (located within 'save_dir').
        use_tbx: defines whether to use tensorboardX writers for logging or not.
        model: model to train; will be uploaded to target device(s) at runtime.
        config: full configuration dictionary of the session; will be incorporated into all saved checkpoints.
        devices: list of (cuda) device IDs to upload the model/tensors to; can be empty if only the CPU is available.
        monitor: specifies the name of the metric that should be monitored on the validation set for model improvement.

    TODO: move static utils to their related modules

    .. seealso::
        | :class:`thelper.train.classif.ImageClassifTrainer`
        | :class:`thelper.train.segm.ImageSegmTrainer`
        | :class:`thelper.train.regr.RegressionTrainer`
        | :func:`thelper.train.utils.create_trainer`
    """

    def __init__(self, session_name, save_dir, model, loaders, config, ckptdata=None):
        # type: (AnyStr, AnyStr, typ.ModelType, typ.MultiLoaderType, typ.ConfigDict, typ.CheckpointContentType) -> None
        """Receives the trainer configuration dictionary, parses it, and sets up the session."""
        if not model or not loaders or not config:
            raise AssertionError("missing input args")
        train_loader, valid_loader, test_loader = loaders
        if not (train_loader or valid_loader or test_loader):
            raise AssertionError("must provide at least one loader with available data")
        if "trainer" not in config or not config["trainer"]:
            raise AssertionError("config missing 'trainer' field")
        trainer_config = config["trainer"]
        packages_log_path = os.path.join(save_dir, "logs", "packages.log")
        with open(packages_log_path, "w") as fd:
            pkgs_list = thelper.utils.get_env_list()
            if pkgs_list:
                for pkg in pkgs_list:
                    fd.write("%s\n" % pkg)
            else:
                fd.write("<n/a>\n")
        train_logger_path = os.path.join(save_dir, "logs", "trainer.log")
        train_logger_format = logging.Formatter("[%(asctime)s - %(process)s] %(levelname)s : %(message)s")
        train_logger_fh = logging.FileHandler(train_logger_path)
        train_logger_fh.setFormatter(train_logger_format)
        self.logger = thelper.utils.get_class_logger()
        self.logger.addHandler(train_logger_fh)
        self.logger.info("created training log for session '%s'" % session_name)
        logstamp = thelper.utils.get_log_stamp()
        repover = thelper.__version__ + ":" + thelper.utils.get_git_stamp()
        self.logger.debug("logstamp = %s" % logstamp)
        self.logger.debug("version = %s" % repover)
        self.name = session_name
        self.train_loader = train_loader
        self.valid_loader = valid_loader
        self.test_loader = test_loader
        self.epochs = 1
        if train_loader:
            if "epochs" not in trainer_config or not trainer_config["epochs"] or int(trainer_config["epochs"]) <= 0:
                raise AssertionError("bad trainer config epoch count")
            self.epochs = int(trainer_config["epochs"])
            if self.epochs <= 0:
                raise AssertionError("should train for at least one epoch")
            # no need to load optimization stuff if not training (i.e. no train_loader)
            # loading optimization stuff later since model needs to be on correct device
            if "optimization" not in trainer_config or not trainer_config["optimization"]:
                raise AssertionError("trainer config missing 'optimization' field")
            self.optimization_config = trainer_config["optimization"]
        else:
            self.logger.info("no training data provided, will run a single epoch on valid/test data")
        self.save_freq = int(trainer_config["save_freq"]) if "save_freq" in trainer_config else 1
        if self.save_freq < 1:
            raise AssertionError("checkpoint save frequency should be integer great or equal to 1")
        self.save_raw = thelper.utils.str2bool(thelper.utils.get_key_def("save_raw", trainer_config, True))
        self.checkpoint_dir = os.path.join(save_dir, "checkpoints")
        if not os.path.exists(self.checkpoint_dir):
            os.mkdir(self.checkpoint_dir)
        output_root_dir = thelper.utils.get_key_def("output_dir", trainer_config, os.path.join(save_dir, "output"))
        output_root_dir = os.path.join(output_root_dir, self.name)
        os.makedirs(output_root_dir, exist_ok=True)
        self.use_tbx = thelper.utils.str2bool(thelper.utils.get_key_def("use_tbx", trainer_config, False))
        if self.use_tbx:
            import tensorboardX
            self.tbx = tensorboardX
            self.logger.debug("tensorboard init : tensorboard --logdir %s --port <your_port>" % output_root_dir)
        self.skip_tbx_histograms = thelper.utils.str2bool(thelper.utils.get_key_def("skip_tbx_histograms", trainer_config, False))
        self.tbx_histogram_freq = thelper.utils.get_key_def("tbx_histogram_freq", trainer_config, 1)
        if self.tbx_histogram_freq < 1:
            raise AssertionError("histogram output frequency should be integer great or equal to 1")
        timestr = time.strftime("%Y%m%d-%H%M%S")
        train_foldername = "train-%s-%s" % (platform.node(), timestr)
        valid_foldername = "valid-%s-%s" % (platform.node(), timestr)
        test_foldername = "test-%s-%s" % (platform.node(), timestr)
        foldernames = [train_foldername, valid_foldername, test_foldername]
        output_paths = [None, None, None]
        for idx, (loader, foldername) in enumerate(zip(loaders, foldernames)):
            if loader:
                out_dir = os.path.join(output_root_dir, foldername)
                if os.path.exists(out_dir):
                    raise AssertionError("output session paths should be unique")
                output_paths[idx] = out_dir
        self.train_output_path, self.valid_output_path, self.test_output_path = output_paths
        self.model = model
        self.config = config
        devices_str = None
        if "device" in trainer_config:
            devices_str = trainer_config["device"]
        elif "train_device" in trainer_config:  # for backw compat only
            devices_str = trainer_config["train_device"]
        self.devices = self._load_devices(devices_str)
        if "metrics" in trainer_config and "base_metrics" in trainer_config:
            raise AssertionError("trainer config should have only one of 'metrics' and 'base_metrics'")
        if ("metrics" in trainer_config and trainer_config["metrics"]) or \
           ("base_metrics" in trainer_config and trainer_config["base_metrics"]):
            if "metrics" in trainer_config:
                self.logger.debug("loading metrics defined in trainer config")
                metrics = thelper.optim.create_metrics(trainer_config["metrics"])
            else:
                self.logger.debug("loading base metrics defined in trainer config")
                metrics = thelper.optim.create_metrics(trainer_config["base_metrics"])
        else:
            metrics = {}
        self.train_metrics = deepcopy(metrics)
        if "train_metrics" in trainer_config and trainer_config["train_metrics"]:
            self.train_metrics = {**self.train_metrics, **thelper.optim.create_metrics(trainer_config["train_metrics"])}
        for metric_name, metric in self.train_metrics.items():
            self.logger.info("parsed train metric '%s': %s" % (metric_name, str(metric)))
        self.valid_metrics = deepcopy(metrics)
        if "valid_metrics" in trainer_config and trainer_config["valid_metrics"]:
            self.valid_metrics = {**self.valid_metrics, **thelper.optim.create_metrics(trainer_config["valid_metrics"])}
        for metric_name, metric in self.valid_metrics.items():
            self.logger.info("parsed valid metric '%s': %s" % (metric_name, str(metric)))
        self.test_metrics = deepcopy(metrics)
        if "test_metrics" in trainer_config and trainer_config["test_metrics"]:
            self.test_metrics = {**self.test_metrics, **thelper.optim.create_metrics(trainer_config["test_metrics"])}
        for metric_name, metric in self.test_metrics.items():
            self.logger.info("parsed test metric '%s': %s" % (metric_name, str(metric)))
        self.monitor, self.monitor_best, self.monitor_best_epoch = None, None, -1
        if "monitor" in trainer_config and trainer_config["monitor"]:
            self.monitor = trainer_config["monitor"]
            if self.monitor not in self.train_metrics and self.monitor not in self.valid_metrics:
                raise AssertionError("metric with name '%s' should be declared in training and/or validation metrics"
                                     % self.monitor)
            if self.monitor in self.valid_metrics:
                self.monitor_goal = self.valid_metrics[self.monitor].goal()
            elif self.monitor in self.train_metrics:
                self.monitor_goal = self.train_metrics[self.monitor].goal()
            if self.monitor_goal == thelper.optim.Metric.minimize:
                self.monitor_best = thelper.optim.Metric.maximize
            elif self.monitor_goal == thelper.optim.Metric.maximize:
                self.monitor_best = thelper.optim.Metric.minimize
            else:
                raise AssertionError("monitored metric does not return proper optimization goal")
        if ckptdata is not None:
            self.monitor_best = ckptdata["monitor_best"]
            self.monitor_best_epoch = ckptdata["monitor_best_epoch"] if "monitor_best_epoch" in ckptdata else -1
            self.optimizer_state = ckptdata["optimizer"] if "optimizer" in ckptdata else None
            self.scheduler_state = ckptdata["scheduler"] if "scheduler" in ckptdata else None
            self.current_iter = ckptdata["iter"]
            self.current_epoch = ckptdata["epoch"]
            self.outputs = ckptdata["outputs"]
        else:
            self.optimizer_state = None
            self.scheduler_state = None
            self.current_iter = 0
            self.current_epoch = 0
            self.outputs = {}
        # callbacks (see ``thelper.typedefs.IterCallbackType`` and ``thelper.typedefs.IterCallbackParams`` definitions)
        self.train_iter_callback = thelper.utils.get_key_def(
            "train_iter_callback", trainer_config, None)    # type: typ.IterCallbackType
        if self.train_iter_callback is not None and isinstance(self.train_iter_callback, str):
            self.train_iter_callback = thelper.utils.import_function(
                self.train_iter_callback)                   # type: typ.IterCallbackType
        self.eval_iter_callback = thelper.utils.get_key_def(
            "eval_iter_callback", trainer_config, None)     # type: typ.IterCallbackType
        if self.eval_iter_callback is not None and isinstance(self.eval_iter_callback, str):
            self.eval_iter_callback = thelper.utils.import_function(
                self.eval_iter_callback)                    # type: typ.IterCallbackType
        display_predictions = thelper.utils.get_key_def("display_preds", trainer_config, False)
        if display_predictions:
            if self.train_iter_callback is not None:
                raise AssertionError("cannot use 'display_preds' while also using an external callback")
            self.train_iter_callback = thelper.utils.import_function("thelper.train.utils._draw_minibatch_wrapper")
            if self.eval_iter_callback is not None:
                raise AssertionError("cannot use 'display_preds' while also using an external callback")
            self.eval_iter_callback = thelper.utils.import_function("thelper.train.utils._draw_minibatch_wrapper")
        if self.train_iter_callback is not None:
            thelper.utils.check_func_signature(self.train_iter_callback, typ.IterCallbackParams)
        if self.eval_iter_callback is not None:
            thelper.utils.check_func_signature(self.eval_iter_callback, typ.IterCallbackParams)
        self.skip_eval_iter = thelper.utils.get_key_def("skip_eval_iter", trainer_config, 0)

    def _init_writer(self, writer, path):
        if self.use_tbx and not writer:
            writer = self.tbx.SummaryWriter(log_dir=path, comment=self.name)
            config_str = json.dumps(self.config, indent=4, sort_keys=False)
            writer.add_text("config", config_str)
            with open(os.path.join(path, "config.json"), "w") as fd:
                fd.write(config_str)
        return writer

    @staticmethod
    def _set_rng_state(seeds, epoch):
        if "torch" in seeds:
            torch.manual_seed(seeds["torch"] + epoch)
            torch.cuda.manual_seed_all(seeds["torch"] + epoch)
        if "numpy" in seeds:
            np.random.seed(seeds["numpy"] + epoch)
        if "random" in seeds:
            random.seed(seeds["random"] + epoch)

    @staticmethod
    def _upload_model(model, dev):
        """Uploads a model to a specific device, wrapping it in ``torch.nn.DataParallel`` if needed."""
        if isinstance(dev, list):
            if len(dev) == 0:
                return model.cpu()
            elif len(dev) == 1:
                return model.cuda(dev[0])
            else:
                return torch.nn.DataParallel(model, device_ids=dev).cuda(dev[0])
        else:
            return model.to(dev)

    @staticmethod
    def _upload_tensor(tensor, dev):
        """Uploads a tensor to a specific device."""
        if isinstance(tensor, list):
            out_tensor = []
            for t in tensor:
                out_tensor.append(Trainer._upload_tensor(t, dev))
            return out_tensor
        elif isinstance(dev, list):
            if len(dev) == 0:
                return tensor.cpu()
            else:
                return tensor.cuda(dev[0])
        else:
            return tensor.to(dev)

    def _load_optimization(self, model, dev):
        """Instantiates and returns all optimization objects required for training the model."""
        config = self.optimization_config  # for abbrev only
        if not isinstance(config, dict):
            raise AssertionError("config should be provided as a dictionary")
        if self.train_loader is None or not self.train_loader:
            raise AssertionError("optimization only useful if training data is available")
        loss = None  # can now be omitted if using custom trainer
        if "loss" in config:
            uploader = functools.partial(self._upload_tensor, dev=dev)
            loss = thelper.optim.create_loss_fn(config["loss"], model, self.train_loader, uploader)
        if "optimizer" not in config or not config["optimizer"]:
            raise AssertionError("optimization config missing 'optimizer' field")
        optimizer = thelper.optim.create_optimizer(config["optimizer"], model)
        scheduler, scheduler_step_metric = None, None
        if "scheduler" in config and config["scheduler"]:
            scheduler, scheduler_step_metric = thelper.optim.create_scheduler(config["scheduler"], optimizer)
        return loss, optimizer, scheduler, scheduler_step_metric

    def _load_devices(self, devices_str=None):
        """Validates and returns the list of CUDA devices available on the system."""
        self.logger.debug("loading available devices")
        available_cuda_devices = thelper.utils.get_available_cuda_devices()
        devices = None
        if devices_str is not None:
            if isinstance(devices_str, str):
                if not devices_str:
                    raise AssertionError("cannot specify empty device name, use default to auto-detect")
                devices_str = devices_str.split(",")
            elif isinstance(devices_str, list):
                if not devices_str:
                    raise AssertionError("cannot specify empty device list, use default to auto-detect")
                if not all([isinstance(dev_str, str) for dev_str in devices_str]):
                    raise AssertionError("unexpected type in device list, should be string")
            else:
                raise AssertionError("unexpected device string type")
            for dev_idx, dev_str in enumerate(devices_str):
                if "cuda" not in dev_str and dev_str != "cpu":
                    raise AssertionError("unknown device type '%s' (expecting 'cpu' or 'cuda:X')" % dev_str)
                elif dev_str == "cpu":
                    if len(devices_str) > 1:
                        raise AssertionError("cannot combine cpu with other devices")
                    return []
                elif dev_str == "cuda" or dev_str == "cuda:all":
                    if len(devices_str) > 1:
                        raise AssertionError("must specify device index (e.g. 'cuda:0') if combining devices")
                    if not available_cuda_devices:
                        raise AssertionError("could not find any available cuda devices")
                    return available_cuda_devices
                elif "cuda:" not in dev_str:
                    raise AssertionError("expecting cuda device format to be 'cuda:X' (where X is device index)")
                cuda_dev_idx = int(dev_str.rsplit(":", 1)[-1])
                if cuda_dev_idx not in available_cuda_devices:
                    raise AssertionError("cuda device '%s' out of range (detected devices = %s)"
                                         % (dev_str, str(available_cuda_devices)))
                if devices is None:
                    devices = [cuda_dev_idx]
                else:
                    devices.append(cuda_dev_idx)
            return devices
        else:
            return available_cuda_devices

    def train(self):
        """Starts the training process.

        This function will train the model until the required number of epochs is reached, and then evaluate it
        on the test data. The setup of loggers, tensorboard writers is done here, so is model improvement tracking
        via monitored metrics. However, the code related to loss computation and back propagation is implemented in
        a derived class via :func:`thelper.train.base.Trainer._train_epoch`.
        """
        if not self.train_loader:
            raise AssertionError("missing training data, invalid loader!")
        self.logger.debug("uploading model to '%s'..." % str(self.devices))
        model = self._upload_model(self.model, self.devices)
        loss, optimizer, scheduler, scheduler_step_metric = self._load_optimization(model, self.devices)
        if self.optimizer_state is not None:
            optimizer.load_state_dict(self.optimizer_state)
            self.optimizer_state = None
        if self.scheduler_state is not None:
            scheduler.load_state_dict(self.scheduler_state)
            self.scheduler_state = None
        self.logger.debug("loss: %s" % str(loss))
        self.logger.debug("optimizer: %s" % str(optimizer))
        latest_loss = math.inf
        train_writer, valid_writer, test_writer = None, None, None
        while self.current_epoch < self.epochs:
            train_writer = self._init_writer(train_writer, self.train_output_path)
            self.logger.info("at epoch#%d for '%s' (dev=%s)" % (self.current_epoch, self.name, str(self.devices)))
            if scheduler:
                if scheduler_step_metric:
                    if scheduler_step_metric == "loss":
                        # todo: use validation loss instead? more stable?
                        scheduler.step(metrics=latest_loss, epoch=self.current_epoch)
                    else:
                        if self.valid_loader and scheduler_step_metric in self.valid_metrics:
                            scheduler.step(metrics=self.valid_metrics[scheduler_step_metric].eval(), epoch=self.current_epoch)
                        elif self.train_loader and scheduler_step_metric in self.train_metrics:
                            scheduler.step(metrics=self.train_metrics[scheduler_step_metric].eval(), epoch=self.current_epoch)
                        else:
                            raise AssertionError("cannot use metric '%s' for scheduler step" % scheduler_step_metric)
                else:
                    scheduler.step(epoch=self.current_epoch)
            if train_writer and not self.skip_tbx_histograms and (self.current_epoch % self.tbx_histogram_freq) == 0:
                for pname, param in model.named_parameters():
                    if "bn" in pname:
                        continue  # skip batch norm modules
                    pname = pname.replace(".", "/")  # for proper grouping
                    if pname.startswith("module/"):
                        pname = pname.replace("module/", "", 1)
                    if pname.startswith("model/"):
                        pname = pname.replace("model/", "", 1)
                    data = param.data.cpu().numpy().flatten()
                    train_writer.add_histogram(pname, data, self.current_epoch)
                    if param.grad is not None:
                        grad = param.grad.data.cpu().numpy().flatten()
                        train_writer.add_histogram(pname + '/grad', grad, self.current_epoch)
            self.logger.debug("learning rate at %.8f" % thelper.optim.get_lr(optimizer))
            self._set_rng_state(self.train_loader.seeds, self.current_epoch)
            model.train()
            for metric in self.train_metrics.values():
                if hasattr(metric, "set_max_accum") and callable(metric.set_max_accum):
                    metric.set_max_accum(len(self.train_loader))  # used to make scalar metric evals smoother between epochs
                if metric.needs_reset():
                    metric.reset()  # if a metric needs to be reset between two epochs, do it here
            if hasattr(self.train_loader, "set_epoch") and callable(self.train_loader.set_epoch):
                self.train_loader.set_epoch(self.current_epoch)
            latest_loss, self.current_iter = self._train_epoch(model, self.current_epoch, self.current_iter, self.devices,
                                                               loss, optimizer, self.train_loader, self.train_metrics,
                                                               self.monitor, train_writer)
            self._write_epoch_output(self.current_epoch, self.train_metrics, train_writer, self.train_output_path,
                                     loss=latest_loss, optimizer=optimizer)
            train_metric_vals = {metric_name: metric.eval() for metric_name, metric in self.train_metrics.items()}
            result = {"train/loss": latest_loss, "train/metrics": train_metric_vals}
            monitor_type_key = "train/metrics"  # if we cannot run validation, will monitor progression on training metrics
            if self.valid_loader:
                self._set_rng_state(self.valid_loader.seeds, self.current_epoch)
                model.eval()
                valid_writer = self._init_writer(valid_writer, self.valid_output_path)
                for metric in self.valid_metrics.values():
                    metric.reset()  # force reset here, we always evaluate from a clean state
                if hasattr(self.valid_loader, "set_epoch") and callable(self.valid_loader.set_epoch):
                    self.valid_loader.set_epoch(self.current_epoch)
                self._eval_epoch(model, self.current_epoch, self.devices, self.valid_loader,
                                 self.valid_metrics, self.monitor, valid_writer)
                self._write_epoch_output(self.current_epoch, self.valid_metrics, valid_writer, self.valid_output_path)
                valid_metric_vals = {metric_name: metric.eval() for metric_name, metric in self.valid_metrics.items()}
                result = {**result, "valid/metrics": valid_metric_vals}
                monitor_type_key = "valid/metrics"  # since validation is available, use that to monitor progression
            new_best = False
            monitor_val = None
            for key, value in result.items():
                if key == monitor_type_key and self.monitor is not None:
                    if self.monitor not in value:
                        raise AssertionError("not monitoring required variable '%s' in metrics" % self.monitor)
                    monitor_val = value[self.monitor]
                    if (self.monitor_goal == thelper.optim.Metric.minimize and monitor_val < self.monitor_best) or \
                       (self.monitor_goal == thelper.optim.Metric.maximize and monitor_val > self.monitor_best):
                        self.monitor_best = monitor_val
                        self.monitor_best_epoch = self.current_epoch
                        new_best = True
                if not isinstance(value, dict):
                    self.logger.debug(" epoch#{} result =>  {}: {}".format(self.current_epoch, str(key), value))
                else:
                    for subkey, subvalue in value.items():
                        self.logger.debug(" epoch#{} result =>  {}:{}: {}".format(self.current_epoch, str(key), str(subkey), subvalue))
            if self.monitor is not None:
                if monitor_val is None:
                    raise AssertionError("training/validation did not produce required monitoring variable '%s'" % self.monitor)
                if new_best:
                    best_str = "(new best value)"
                else:
                    best_str = ("(previous best = %s @ epoch = %d)" % (self.monitor_best, self.monitor_best_epoch))
                self.logger.info("epoch %d, monitored %s = %s  %s" % (self.current_epoch, self.monitor, monitor_val, best_str))
            self.outputs[self.current_epoch] = result
            if new_best or (self.current_epoch % self.save_freq) == 0:
                self.logger.info("saving checkpoint @ epoch#%d" % self.current_epoch)
                self._save(self.current_epoch, self.current_iter, optimizer, scheduler, save_best=new_best)
            self.current_epoch += 1
        self.logger.info("training for session '%s' done" % self.name)
        if self.test_loader:
            # reload 'best' model checkpoint on cpu (will remap to current device setup)
            filename_best = os.path.join(self.checkpoint_dir, "ckpt.best.pth")
            self.logger.info("loading best model & initializing final test run")
            ckptdata = thelper.utils.load_checkpoint(filename_best, map_location="cpu")
            if self.config != ckptdata["config"]:  # todo: dig into members and check only critical ones
                raise AssertionError("could not load compatible best checkpoint to run test eval")
            if self.save_raw:
                self.model.load_state_dict(ckptdata["model"])
            else:
                self.model = ckptdata["model"]
            best_epoch = ckptdata["epoch"]
            model = self._upload_model(self.model, self.devices)
            self._set_rng_state(self.test_loader.seeds, best_epoch)
            model.eval()
            test_writer = self._init_writer(test_writer, self.test_output_path)
            for metric in self.test_metrics.values():
                metric.reset()  # force reset here, we always evaluate from a clean state
            if hasattr(self.test_loader, "set_epoch") and callable(self.test_loader.set_epoch):
                self.test_loader.set_epoch(best_epoch)
            self._eval_epoch(model, best_epoch, self.devices, self.test_loader,
                             self.test_metrics, self.monitor, test_writer)
            self._write_epoch_output(best_epoch, self.test_metrics, test_writer, self.test_output_path)
            test_metric_vals = {metric_name: metric.eval() for metric_name, metric in self.test_metrics.items()}
            self.outputs[best_epoch] = {**ckptdata["outputs"], "test/metrics": test_metric_vals}
            for key, value in self.outputs[best_epoch].items():
                if not isinstance(value, dict):
                    self.logger.debug(" final result =>  {}: {}".format(str(key), value))
                else:
                    for subkey, subvalue in value.items():
                        self.logger.debug(" final result =>  {}:{}: {}".format(str(key), str(subkey), subvalue))
            self.logger.info("evaluation for session '%s' done" % self.name)
        return self.outputs

    def eval(self):
        """Starts the evaluation process.

        This function will evaluate the model using the test data (or the validation data, if no test data is available),
        and return the results. Note that the code related to the forwarding of samples inside the model itself is implemented
        in a derived class via :func:`thelper.train.base.Trainer._eval_epoch`.
        """
        if not self.valid_loader and not self.test_loader:
            raise AssertionError("missing validation/test data, invalid loaders!")
        self.logger.debug("uploading model to '%s'..." % str(self.devices))
        model = self._upload_model(self.model, self.devices)
        result = {}
        valid_writer, test_writer = None, None
        if self.test_loader:
            self._set_rng_state(self.test_loader.seeds, self.current_epoch)
            model.eval()
            test_writer = self._init_writer(test_writer, self.test_output_path)
            for metric in self.test_metrics.values():
                metric.reset()  # force reset here, we always evaluate from a clean state
            if hasattr(self.test_loader, "set_epoch") and callable(self.test_loader.set_epoch):
                self.test_loader.set_epoch(self.current_epoch)
            self._eval_epoch(model, self.current_epoch, self.devices, self.test_loader,
                             self.test_metrics, self.monitor, test_writer)
            self._write_epoch_output(self.current_epoch, self.test_metrics, test_writer, self.test_output_path)
            test_metric_vals = {metric_name: metric.eval() for metric_name, metric in self.test_metrics.items()}
            result = {**result, **test_metric_vals}
        elif self.valid_loader:
            self._set_rng_state(self.valid_loader.seeds, self.current_epoch)
            model.eval()
            valid_writer = self._init_writer(valid_writer, self.valid_output_path)
            for metric in self.valid_metrics.values():
                metric.reset()  # force reset here, we always evaluate from a clean state
            if hasattr(self.valid_loader, "set_epoch") and callable(self.valid_loader.set_epoch):
                self.valid_loader.set_epoch(self.current_epoch)
            self._eval_epoch(model, self.current_epoch, self.devices, self.valid_loader,
                             self.valid_metrics, self.monitor, valid_writer)
            self._write_epoch_output(self.current_epoch, self.valid_metrics, valid_writer, self.valid_output_path)
            valid_metric_vals = {metric_name: metric.eval() for metric_name, metric in self.valid_metrics.items()}
            result = {**result, **valid_metric_vals}
        for key, value in result.items():
            if not isinstance(value, dict):
                self.logger.debug(" final result =>  {}: {}".format(str(key), value))
            else:
                for subkey, subvalue in value.items():
                    self.logger.debug(" final result =>  {}:{}: {}".format(str(key), str(subkey), subvalue))
        self.outputs[self.current_epoch] = result
        self.logger.info("evaluation for session '%s' done" % self.name)
        return result

    @abstractmethod
    def _train_epoch(self, model, epoch, iter, dev, loss, optimizer, loader, metrics, monitor=None, writer=None):
        """Trains the model for a single epoch using the provided objects.

        Args:
            model: the model to train that is already uploaded to the target device(s).
            epoch: the epoch index we are training for (0-based).
            iter: the iteration count at the start of the current epoch.
            dev: the target device that tensors should be uploaded to.
            loss: the loss function used to evaluate model fidelity.
            optimizer: the optimizer used for back propagation.
            loader: the data loader used to get transformed training samples.
            metrics: the dictionary of metrics to update every iteration.
            monitor: name of the metric to update/monitor for improvements.
            writer: the writer used to store tbx events/messages/metrics.
        """
        raise NotImplementedError

    @abstractmethod
    def _eval_epoch(self, model, epoch, dev, loader, metrics, monitor=None, writer=None):
        """Evaluates the model using the provided objects.

        Args:
            model: the model to evaluate that is already uploaded to the target device(s).
            epoch: the epoch index we are training for (0-based).
            dev: the target device that tensors should be uploaded to.
            loader: the data loader used to get transformed valid/test samples.
            metrics: the dictionary of metrics to update every iteration.
            monitor: name of the metric to update/monitor for improvements.
            writer: the writer used to store tbx events/messages/metrics.
        """
        raise NotImplementedError

    def _write_epoch_output(self, epoch, metrics, tbx_writer, output_path, loss=None, optimizer=None):
        """Writes the cumulative evaluation result of all metrics using a specific writer."""
        self.logger.debug("writing epoch metrics to '%s'" % output_path)
        if not os.path.exists(output_path):
            os.makedirs(output_path)
        if tbx_writer is not None and loss is not None and optimizer is not None:
            tbx_writer.add_scalar("epoch/loss", loss, epoch)
            tbx_writer.add_scalar("epoch/lr", thelper.optim.get_lr(optimizer), epoch)
        for metric_name, metric in metrics.items():
            if metric.is_scalar():
                if tbx_writer is not None:
                    tbx_writer.add_scalar("epoch/%s" % metric_name, metric.eval(), epoch)
            if hasattr(metric, "render") and callable(metric.render):
                img = metric.render()
                if img is not None:
                    if tbx_writer is not None:
                        tbx_writer.add_image(metric_name, img, epoch, dataformats="HWC")
                    raw_filename = "%s-%04d.png" % (metric_name, epoch)
                    raw_filepath = os.path.join(output_path, raw_filename)
                    cv.imwrite(raw_filepath, img[..., [2, 1, 0]])
            txt = metric.print() if hasattr(metric, "print") and callable(metric.print) else None
            if not txt:
                eval_res = metric.eval()
                if eval_res is not None:
                    if isinstance(eval_res, float):
                        txt = "%.4f" % eval_res  # make sure we always have decent precision
                    else:
                        txt = str(eval_res)
            if txt:
                raw_filename = "%s-%04d.txt" % (metric_name, epoch)
                raw_filepath = os.path.join(output_path, raw_filename)
                with open(raw_filepath, "w") as fd:
                    fd.write(txt)

    def _save(self, epoch, iter, optimizer, scheduler, save_best=False):
        """Saves a session checkpoint containing all the information required to resume training."""
        # logically, this should only be called during training (i.e. with a valid optimizer)
        log_stamp = thelper.utils.get_log_stamp()
        curr_state = {
            "name": self.name,
            "epoch": epoch,
            "iter": iter,
            "source": log_stamp,
            "sha1": thelper.utils.get_git_stamp(),
            "version": thelper.__version__,
            "task": str(self.model.task) if self.save_raw else self.model.task,
            "outputs": self.outputs[epoch],
            # we save model type/params here in case those are not in the current config
            "model": self.model.state_dict() if self.save_raw else self.model,
            "model_type": self.model.get_name(),
            "model_params": self.model.config if self.model.config else {},
            "optimizer": optimizer.state_dict() if optimizer is not None else None,
            "scheduler": scheduler.state_dict() if (scheduler is not None and
                                                    hasattr(scheduler, "state_dict")) else None,
            "monitor_best": self.monitor_best,
            "monitor_best_epoch": self.monitor_best_epoch,
            "config": self.config  # note: this is the global app config
        }
        filename = "ckpt.%04d.%s.pth" % (epoch, log_stamp)
        filename = os.path.join(self.checkpoint_dir, filename)
        self.logger.debug("writing checkpoint to '%s'" % filename)
        torch.save(curr_state, filename)
        if save_best:
            filename_best = os.path.join(self.checkpoint_dir, "ckpt.best.pth")
            self.logger.debug("writing checkpoint to '%s'" % filename_best)
            torch.save(curr_state, filename_best)



"""Regression trainer/evaluator implementation module."""
import logging

import numpy as np
import torch
import torch.optim

import thelper.utils
from thelper.train.base import Trainer

logger = logging.getLogger(__name__)


class RegressionTrainer(Trainer):
    """Trainer interface specialized for generic (n-dim) regression.

    This class implements the abstract functions of :class:`thelper.train.base.Trainer` required to train/evaluate
    a model for generic regression (i.e. n-dim target value prediction). It also provides a utility function
    for fetching i/o packets (input tensors, target values) from a sample, and that converts those into tensors for
    forwarding and loss estimation.

    .. seealso::
        | :class:`thelper.train.base.Trainer`
    """

    def __init__(self, session_name, save_dir, model, loaders, config, ckptdata=None):
        """Receives session parameters, parses tensor/target keys from task object, and sets up metrics."""
        super().__init__(session_name, save_dir, model, loaders, config, ckptdata=ckptdata)
        if not isinstance(self.model.task, thelper.tasks.Regression):
            raise AssertionError("expected task to be regression")
        self.input_key = self.model.task.get_input_key()
        self.target_key = self.model.task.get_gt_key()
        self.input_shape = self.model.task.get_input_shape()
        self.target_shape = self.model.task.get_target_shape()
        self.target_type = self.model.task.get_target_type()
        self.target_min = self.model.task.get_target_min()
        if isinstance(self.target_min, np.ndarray):
            self.target_min = torch.from_numpy(self.target_min)
        self.target_max = self.model.task.get_target_max()
        if isinstance(self.target_max, np.ndarray):
            self.target_max = torch.from_numpy(self.target_max)
        self.meta_keys = self.model.task.get_meta_keys()

    def _to_tensor(self, sample):
        """Fetches and returns tensors of inputs and targets from a batched sample dictionary."""
        if not isinstance(sample, dict):
            raise AssertionError("trainer expects samples to come in dicts for key-based usage")
        if self.input_key not in sample:
            raise AssertionError("could not find input key '%s' in sample dict" % self.input_key)
        input_val = sample[self.input_key]
        if isinstance(input_val, np.ndarray):
            input_val = torch.from_numpy(input_val)
        if not isinstance(input_val, torch.Tensor):
            raise AssertionError("unexpected input type; should be torch.Tensor")
        if self.input_shape is not None:
            if input_val.dim() != len(self.input_shape) + 1:
                raise AssertionError("expected input as Nx[shape] where N = batch size")
            if self.input_shape != input_val.shape[1:]:
                raise AssertionError("invalid input shape; got '%s', expected '%s'" % (input_val.shape[1:], self.input_shape))
        target = None
        if self.target_key in sample:
            target = sample[self.target_key]
            if isinstance(target, np.ndarray):
                if self.target_type is not None and target.dtype != self.target_type:
                    raise AssertionError("unexpected target type, should be %s" % str(self.target_type))
                target = torch.from_numpy(target)
            if not isinstance(target, torch.Tensor):
                raise AssertionError("unexpected target type; should be torch.Tensor")
            if self.target_shape is not None:
                if target.dim() != len(self.target_shape) + 1:
                    raise AssertionError("expected target as Nx[shape] where N = batch size")
                if self.target_shape != target.shape[1:]:
                    raise AssertionError("invalid target shape; got '%s', expected '%s'" % (target.shape[1:], self.target_shape))
        return input_val, target

    def _train_epoch(self, model, epoch, iter, dev, loss, optimizer, loader, metrics, monitor=None, writer=None):
        """Trains the model for a single epoch using the provided objects.

        Args:
            model: the model to train that is already uploaded to the target device(s).
            epoch: the epoch index we are training for (0-based).
            iter: the iteration count at the start of the current epoch.
            dev: the target device that tensors should be uploaded to.
            loss: the loss function used to evaluate model fidelity.
            optimizer: the optimizer used for back propagation.
            loader: the data loader used to get transformed training samples.
            metrics: the list of metrics to evaluate after every iteration.
            monitor: name of the metric to update/monitor for improvements.
            writer: the writer used to store tbx events/messages/metrics.
        """
        if not loss:
            raise AssertionError("missing loss function")
        if not optimizer:
            raise AssertionError("missing optimizer")
        if not loader:
            raise AssertionError("no available data to load")
        if not isinstance(metrics, dict):
            raise AssertionError("expect metrics as dict object")
        epoch_loss = 0
        epoch_size = len(loader)
        self.logger.debug("fetching data loader samples...")
        for idx, sample in enumerate(loader):
            input_val, target = self._to_tensor(sample)
            # todo: add support to fraction samples that are too big for a single iteration
            # (e.g. when batching non-image data that would be too inefficient one sample at a time)
            optimizer.zero_grad()
            if target is None:
                raise AssertionError("groundtruth required when training a model")
            if isinstance(input_val, list):
                raise AssertionError("missing regr trainer support for augmented minibatches")  # todo
            target = self._upload_tensor(target, dev)
            iter_pred = model(self._upload_tensor(input_val, dev))
            iter_loss = loss(iter_pred, target.float())
            iter_loss.backward()
            if metrics:
                meta = {key: sample[key] if key in sample else None for key in self.meta_keys} if self.meta_keys else None
                for metric in metrics.values():
                    metric.accumulate(iter_pred.detach().cpu(), target.detach().cpu(), meta=meta)
            if self.train_iter_callback is not None:
                self.train_iter_callback(sample=sample, task=self.model.task, pred=iter_pred,
                                         iter_idx=iter, max_iters=epoch_size,
                                         epoch_idx=epoch, max_epochs=self.epochs)
            epoch_loss += iter_loss.item()
            optimizer.step()
            monitor_output = ""
            if monitor is not None and monitor in metrics:
                monitor_output = "   {}: {:.2f}".format(monitor, metrics[monitor].eval())
            self.logger.info(
                "train epoch#{}  (iter#{})   batch: {}/{} ({:.0f}%)   loss: {:.6f}{}".format(
                    epoch,
                    iter,
                    idx + 1,
                    epoch_size,
                    ((idx + 1) / epoch_size) * 100.0,
                    iter_loss.item(),
                    monitor_output
                )
            )
            if writer:
                writer.add_scalar("iter/loss", iter_loss.item(), iter)
                for metric_name, metric in metrics.items():
                    if metric.is_scalar():  # only useful assuming that scalar metrics are smoothed...
                        writer.add_scalar("iter/%s" % metric_name, metric.eval(), iter)
            iter += 1
        epoch_loss /= epoch_size
        return epoch_loss, iter

    def _eval_epoch(self, model, epoch, dev, loader, metrics, monitor=None, writer=None):
        """Evaluates the model using the provided objects.

        Args:
            model: the model to evaluate that is already uploaded to the target device(s).
            epoch: the epoch number we are evaluating for (0-based).
            dev: the target device that tensors should be uploaded to.
            loader: the data loader used to get transformed valid/test samples.
            metrics: the dictionary of metrics to update every iteration.
            monitor: name of the metric to update/monitor for improvements.
            writer: the writer used to store tbx events/messages/metrics.
        """
        if not loader:
            raise AssertionError("no available data to load")
        with torch.no_grad():
            epoch_size = len(loader)
            self.logger.debug("fetching data loader samples...")
            for idx, sample in enumerate(loader):
                if idx < self.skip_eval_iter:
                    continue  # skip until previous iter count (if set externally; no effect otherwise)
                input_val, target = self._to_tensor(sample)
                if isinstance(input_val, list):
                    raise AssertionError("missing regr trainer support for augmented minibatches")  # todo
                pred = model(self._upload_tensor(input_val, dev))
                if metrics:
                    meta = {key: sample[key] if key in sample else None
                            for key in self.meta_keys} if self.meta_keys else None
                    for metric in metrics.values():
                        metric.accumulate(pred.cpu(), target.cpu() if target is not None else None, meta=meta)
                if self.eval_iter_callback is not None:
                    self.eval_iter_callback(sample=sample, task=self.model.task, pred=pred,
                                            iter_idx=idx, max_iters=epoch_size,
                                            epoch_idx=epoch, max_epochs=self.epochs)
                self.logger.info(
                    "eval epoch#{}   batch: {}/{} ({:.0f}%){}".format(
                        epoch,
                        idx + 1,
                        epoch_size,
                        ((idx + 1) / epoch_size) * 100.0,
                        "   {}: {:.2f}".format(monitor, metrics[monitor].eval()) if monitor is not None else ""
                    )
                )



import unittest

from torch.tensor import Tensor

from thelper.optim.metrics import RawPredictions


class TestMetrics(unittest.TestCase):
    def test_raw_predictions_metric_basic_use_case(self):
        """Validations for basic use case:
            - predictions are gradually saved on 'accumulate' calls
            - non-serializable tensors are converted to JSON lists
            - meta information are transferred accordingly to matching predictions
        """

        metric = RawPredictions()
        list1d = [1., 2., 3., 4., 5., 6.]
        list2d = [[1., 2.], [3., 4.], [5., 6.]]
        list3d = [[[1.], [2.]], [[3.], [4.]], [[5.], [6.]]]

        gt = ['class1', 'class2']
        pred = Tensor(list2d)   # predictions of sample batch
        meta = {
            'l1d': list1d,
            'l2d': list2d,
            'l3d': list3d,
        }

        # should accumulate and convert to serializable types
        metric.accumulate(pred, gt, meta)
        assert isinstance(metric.predictions, list)
        assert len(metric.predictions) == len(pred), "saved predictions should match dimensions"
        for mp, tp in zip(metric.predictions, pred):
            assert isinstance(mp, dict), "predictions should be a dictionary of information"
            assert isinstance(mp['predictions'], list), "predictions tensor should be converted to serializable list"
            for mp_pred, tp_pred in zip(mp['predictions'], tp):
                assert mp_pred == tp_pred, "corresponding predictions should be transferred to serializable list"
            for k in meta.keys():
                assert k in mp, "all meta info should be added"
                assert isinstance(mp[k], list), "no tensors should be added (non-serializable)"

        # second call appends more predictions
        metric.accumulate(pred, gt, meta)
        assert len(metric.predictions) == len(pred) * 2, "saved predictions should be extended with new ones"
        for mp, tp in zip(metric.predictions, list(pred) + list(pred)):
            for mp_pred, tp_pred in zip(mp['predictions'], tp):
                assert mp_pred == tp_pred, "corresponding predictions should be transferred to serializable list"


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

import glob
import io
import os
import re

import setuptools
from setuptools.command.build_ext import build_ext


def read(*names, **kwargs):
    return io.open(
        os.path.join(os.path.dirname(__file__), *names),
        encoding=kwargs.get("encoding", "utf8")
    ).read()


# Enable code coverage for C code: we can't use CFLAGS=-coverage in tox.ini, since that may mess with compiling
# dependencies (e.g. numpy). Therefore we set SETUPPY_CFLAGS=-coverage in tox.ini and copy it to CFLAGS here (after
# deps have been safely installed).
if "TOXENV" in os.environ and "SETUPPY_CFLAGS" in os.environ:
    os.environ["CFLAGS"] = os.environ["SETUPPY_CFLAGS"]


class OptionalBuildExt(build_ext):
    """Allow the building of C extensions to fail."""

    def run(self):
        try:
            build_ext.run(self)
        except Exception as e:
            self._unavailable(e)
            self.extensions = []  # avoid copying missing files (it would fail).

    def _unavailable(self, e):
        print("*" * 80)
        print("""WARNING:

    An optional code optimization (C extension) could not be compiled.

    Optimizations for this package will not be available!
        """)
        print("CAUSE:")
        print("")
        print("    " + repr(e))
        print("*" * 80)


on_rtd = os.environ.get('READTHEDOCS') == 'True'

with open("requirements.txt") as reqfd:
    install_requires = reqfd.read()

setuptools.setup(
    name="thelper",
    version="0.2.8",
    license="Apache Software License 2.0",
    description="Training framework & tools for PyTorch-based machine learning projects.",
    long_description="%s\n%s" % (
        re.compile("^.. start-badges.*^.. end-badges", re.M | re.S).sub("", read("README.rst")),
        re.sub(":[a-z]+:`~?(.*?)`", r"``\1``", read("CHANGELOG.rst"))
    ),
    author="Pierre-Luc St-Charles",
    author_email="stcharpl@crim.ca",
    url="https://github.com/plstcharles/thelper",
    packages=setuptools.find_packages(exclude=("test",)),
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: Unix",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",  # we assume dict insert order will be kept intact... (>=3.6)
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    keywords=["pytorch", "trainer", "loader"],
    install_requires=install_requires if not on_rtd else [],  # bypass deps install on rtd
    python_requires="~=3.6",
    extras_require={
        "rst": ["docutils>=0.11"],
    },
    entry_points={
        "console_scripts": [
            "thelper = thelper.cli:main",
        ]
    },
    cmdclass={"build_ext": OptionalBuildExt},
    ext_modules=[
        setuptools.Extension(
            os.path.splitext(os.path.relpath(path, "src").replace(os.sep, "."))[0],
            sources=[path],
            include_dirs=[os.path.dirname(path)]
        )
        for root, _, _ in os.walk("src")
        for path in glob.glob(os.path.join(root, "*.c"))
    ],
)



"""Task utility functions & base interface module.

This module contains utility functions used to instantiate tasks and check their compatibility,
and the base interface used to define new tasks.
"""

import logging

import thelper.utils

logger = logging.getLogger(__name__)


def create_task(config):
    """Parses a configuration dictionary or repr string and instantiates a task from it.

    If a string is provided, it will first be parsed to get the task type, and then the object will be
    instantiated by forwarding the parameters contained in the string to the constructor of that
    type. Note that it is important for this function to work that the constructor argument names
    match the names of parameters printed in the task's ``__repr__`` function.

    If a dict is provided, it should contain a 'type' and a 'params' field with the values required
    for direct instantiation.

    .. seealso::
        | :class:`thelper.tasks.utils.Task`
        | :func:`thelper.tasks.utils.Task.__repr__`
    """
    if config is None or not isinstance(config, (str, dict)):
        raise AssertionError("unexpected config type (should be str or dict)")
    if isinstance(config, dict):
        if "type" not in config or not isinstance(config["type"], str):
            raise AssertionError("invalid field 'type' in task config")
        task_type = thelper.utils.import_class(config["type"])
        task_params = thelper.utils.get_key(["params", "parameters"], config)
        if not isinstance(task_params, dict):
            raise AssertionError("invalid field 'params' in task config")
        task = task_type(**task_params)
        if not isinstance(task, thelper.tasks.Task):
            raise AssertionError("the task must be derived from 'thelper.tasks.Task'")
        return task
    elif isinstance(config, str):
        task_type_name = config.split(": ")[0]
        if "." not in task_type_name:
            # dirty hotfix
            task_type_name = "thelper.tasks." + task_type_name
        task_type = thelper.utils.import_class(task_type_name)
        task_params = eval(": ".join(config.split(": ")[1:]))
        task = task_type(**task_params)
        if not isinstance(task, thelper.tasks.Task):
            raise AssertionError("the task must be derived from 'thelper.tasks.Task'")
        return task


def create_global_task(tasks):
    """Returns a new task object that is compatible with a list of subtasks.

    When different datasets must be combined in a session, the tasks they define must also be
    merged. This functions allows us to do so as long as the tasks all share a common objective.
    If creating a globally-compatible task is impossible, this function will raise an exception.
    Otherwise, the returned task object can be used to replace the subtasks of all used datasets.

    .. seealso::
        | :class:`thelper.tasks.utils.Task`
        | :func:`thelper.tasks.utils.create_task`
        | :func:`thelper.data.utils.create_parsers`
    """
    if tasks is None:
        return None
    if not isinstance(tasks, list):
        raise AssertionError("tasks should be provided as list")
    ref_task = None
    for task in tasks:
        if task is None:
            # skip all undefined tasks
            continue
        if not isinstance(task, thelper.tasks.Task):
            raise AssertionError("all tasks should derive from thelper.tasks.Task")
        if ref_task is None:
            # no reference task set; take the first instance and continue to next
            ref_task = task
            continue
        if type(ref_task) != Task:
            # reference task already specialized, we can ask it for compatible instances
            ref_task = ref_task.get_compat(task)
        else:
            # otherwise, keep asking the new one to stay compatible with the base ref
            ref_task = task.get_compat(ref_task)
    return ref_task


class Task(object):
    """Basic task interface that defines a training objective and that holds sample i/o keys.

    Since the framework's data loaders expect samples to be passed in as dictionaries, keys
    are required to obtain the input that should be forwarded to a model, and to obtain the
    groundtruth required for the evaluation of model predictions. Other keys might also be
    kept by this interface for reference (these are considered meta keys).

    Note that while this interface can be instantiated directly, trainers and models might
    not be provided enough information about their goal to be corectly instantiated. Thus,
    specialized task objects derived from this base class should be used if possible.

    Attributes:
        input_key: the key used to fetch input tensors from a sample dictionary.
        gt_key: the key used to fetch gt tensors from a sample dictionary.
        meta_keys: the list of extra keys provided by the data parser inside each sample.

    .. seealso::
        | :class:`thelper.tasks.classif.Classification`
        | :class:`thelper.tasks.segm.Segmentation`
        | :class:`thelper.tasks.regr.Regression`
    """

    def __init__(self, input_key, gt_key=None, meta_keys=None):
        """Receives and stores the input tensor key, the groundtruth tensor key, and the extra
        (meta) keys produced by the dataset parser(s).
        """
        if input_key is None:
            raise AssertionError("task input key cannot be None (input tensors must always be available)")
        self.input_key = input_key
        self.gt_key = gt_key
        self.meta_keys = []
        if meta_keys is not None:
            if not isinstance(meta_keys, list):
                raise AssertionError("meta keys should be provided as a list")
            self.meta_keys = meta_keys

    def get_input_key(self):
        """Returns the key used to fetch input data tensors from a sample dictionary.

        The key can be of any type, as long as it can be used to index a dictionary. Print-
        friendly types (e.g. string) are recommended for debugging. This key can never be
        ``None``, as input tensors should always be available in loaded samples.
        """
        return self.input_key

    def get_gt_key(self):
        """Returns the key used to fetch groundtruth data tensors from a sample dictionary.

        The key can be of any type, as long as it can be used to index a dictionary. Print-
        friendly types (e.g. string) are recommended for debugging. If groundtruth is not
        available through the dataset parsers, this function should teturn ``None``.
        """
        return self.gt_key

    def get_meta_keys(self):
        """Returns a list of keys used to carry metadata and auxiliary info in samples.

        The keys can be of any type, as long as they can be used to index a dictionary.
        Print-friendly types (e.g. string) are recommended for debugging. This list can
        be empty if the dataset/model does not provide/require any extra inputs.
        """
        return self.meta_keys

    def check_compat(self, other, exact=False):
        """Returns whether the current task is compatible with the provided one or not.

        This is useful for sanity-checking, and to see if the inputs/outputs of two models
        are compatible. It should be overridden in derived classes to specialize the
        compatibility verification. If ``exact = True``, all fields will be checked for
        exact (perfect) compatibility.
        """
        if type(other) == Task:
            return (self.get_input_key() == other.get_input_key() and
                    self.get_gt_key() == other.get_gt_key())
        return False

    def get_compat(self, other):
        """Returns a task instance compatible with the current task and the given one."""
        if type(other) == Task:
            if not self.check_compat(other):
                raise AssertionError("cannot create compatible task instance between:\n"
                                     "\tself: %s\n\tother: %s" % (str(self), str(other)))
            meta_keys = list(set(self.get_meta_keys() + other.get_meta_keys()))
            return Task(self.get_input_key(), self.get_gt_key(), meta_keys)
        else:
            raise AssertionError("cannot combine task type '%s' with '%s'" % (str(other.__class__), str(self.__class__)))

    def __repr__(self):
        """Creates a print-friendly representation of an abstract task.

        Note that this representation might also be used to check the compatibility of tasks
        without importing the whole framework. Therefore, it should contain all the necessary
        information about the task. The name of the parameters herein should also match the
        argument names given to the constructor in case we need to recreate a task object from
        this string.
        """
        return self.__class__.__module__ + "." + self.__class__.__qualname__ + ": " + str({
            "input_key": self.get_input_key(),
            "gt_key": self.get_gt_key(),
            "meta_keys": self.get_meta_keys()
        })



"""Trainer package.

This package contains classes specialized for training models on various tasks.
"""

import logging

from thelper.train.base import Trainer  # noqa: F401
from thelper.train.classif import ImageClassifTrainer  # noqa: F401
from thelper.train.regr import RegressionTrainer  # noqa: F401
from thelper.train.segm import ImageSegmTrainer  # noqa: F401
from thelper.train.utils import create_trainer  # noqa: F401

logger = logging.getLogger("thelper.train")
